class MissingRunDevModule(ValueError):
    def __init__(self, app):
        super().__init__(
            f"""
The app {app.shell_app_id()} does not define the `RUN_DEV_MODULE` attribute.

Your app must be runnable with this command:
    `python -m your_app_package.your_show_dev_module`
Or if you're using uv:
    `uv run python -m your_app_package.your_show_dev_module`

You must implement "your_show_**dev**_module" whith this content:

```python

from your_app_package import YourAppClass

if __name__ in {"__main__", "__mp_main__"}:
    YourAppClass().run_app(native=False, reload=True)
```

"""
        )


class MissingRunNativeModule(ValueError):
    def __init__(self, app):
        super().__init__(
            f"""
The app {app.shell_app_id()} does not define the `RUN_NATIVE_MODULE` attribute.

Your app must be runnable with this command:
    `python -m your_app_package.your_show_nativ_module`
Or if you're using uv:
    `uv run python -m your_app_package.your_show_native_module`

You must implement "your_show_**native**_module" whith this content:

```python

from your_app_package import YourAppClass

if __name__ == "__main__":
    YourAppClass().run_app(native=True, reload=False)
```
"""
        )



from __future__ import annotations
from typing import TYPE_CHECKING, Literal

import os
from pathlib import Path
import platform
import tempfile
import importlib.metadata

from .base_config import BaseConfig, SettingsConfigDict, Field
from .project import Project

from tgzr.package_management.package_manager import PackageManager, Venv

if TYPE_CHECKING:
    from .workspace import Workspace


class StudioConfig(BaseConfig):
    model_config = SettingsConfigDict(env_prefix="tgzr_studio_defaults_")

    default_index: str | None = Field(
        None,
        description="Default package index for this studio. Use None to fallback to Work config",
    )


class Studio:

    @classmethod
    def config_path(cls, workspace: Workspace, name: str) -> Path:
        return workspace.path / name / ".tgzr_studio"

    def __init__(self, workspace: Workspace, name: str) -> None:
        self._workspace = workspace
        self._name = name
        self._config_path = self.config_path(workspace, name)
        self._config = StudioConfig.from_file(self._config_path)

        self._venvs_path = self.path / "venvs"
        self._tgzr_venv_group = "_"
        self._tgzr_venv_name = "tgzr"
        self._projects_venv_group = "projects"

    @property
    def config(self) -> StudioConfig:
        return self._config

    @property
    def name(self) -> str:
        return self._name

    @property
    def workspace(self) -> Workspace:
        return self._workspace

    @property
    def path(self) -> Path:
        return self.workspace.path / self.name

    def resolve_index(self, index: str | None) -> str | None:
        """
        Returns the path of an index, making it absolute in case
        of relative local path.

        If index is None, the default index will be returned.
        If no default index in found on the Studio nor on the Workspace,
        None is returned.
        """
        index = index or self.config.default_index
        return self._workspace.resolve_index(index, relative_to=self.path)

    def save_config(self) -> Path:
        """Returns the path of the saved file."""
        self._config.to_file(
            self._config_path,
            allow_overwrite=True,
            header_text="tgzr studio config",
        )
        return self._config_path

    def exists(self) -> bool:
        return self.path.exists()

    def ensure_exists(self):
        if not self.exists:
            self.create()

    @property
    def projects_venv_path(self) -> Path:
        return self._venvs_path / self._projects_venv_group

    def project_venv_path(self, project_name: str) -> Path:
        return self.projects_venv_path / project_name

    def get_project_names(self) -> list[str]:
        # TODO: filter by looking for the project config file
        if not self.projects_venv_path.is_dir():
            return []
        return os.listdir(self.projects_venv_path)

    def get_project(self, project_name: str) -> Project:
        return Project(self, project_name)

    def create(
        self,
        index: str | None = None,
        find_links: str | None = None,
        allow_prerelease: bool = False,
    ):
        """
        Create (or update) this Studio and seed it with the `tgzr.shell` package.
        If `index` is None, the studio's default index is used.
        """
        self.path.mkdir(exist_ok=True)
        self.save_config()

        pm = PackageManager(self._venvs_path)
        venv = pm.create_venv(
            self._tgzr_venv_name,
            group=self._tgzr_venv_group,
            exist_ok=True,
            prompt=self.name,
        )

        index = self.resolve_index(index)
        ok = venv.install_packages(
            "tgzr.shell tgzr.shell_apps.manager_panel",
            index=index,
            find_links=find_links,
            allow_prerelease=allow_prerelease,
            update=True,
        )
        if not ok:
            raise Exception("Could not install tgzr in Studio :( Aborting.")

        tgzr_exe = venv.get_exe("tgzr")
        tgzr_shortcut = self.path / "tgzr"
        if tgzr_shortcut.exists():
            tgzr_shortcut.unlink()
        pm.create_shortcut(tgzr_exe, tgzr_shortcut, relative=True)

    def create_project(
        self,
        project_name: str,
        required_packages: list[str] = [],
        index: str | None = None,
        find_links: str | None = None,
        allow_prerelease: bool = False,
    ) -> tuple[bool, list[str]]:
        """
        Create (or update) this project and seed it with the `tgzr.shell` package.
        If `index` is None, the studio's default index is used.
        All requirements in `required_packages` are installed two.

        Returns (all_requirement_install_ok, list_of_failed_install)

        """
        self.path.mkdir(exist_ok=True)
        self.save_config()

        pm = PackageManager(self._venvs_path)
        venv = pm.create_venv(
            project_name,
            group=self._projects_venv_group,
            exist_ok=True,
            prompt=f"{self.name}:{project_name}",
        )

        index = self.resolve_index(index)
        ok = venv.install_packages(
            "tgzr.shell tgzr.shell_apps.manager_panel",
            index=index,
            find_links=find_links,
            allow_prerelease=allow_prerelease,
            update=True,
        )
        if not ok:
            raise Exception("Could not install tgzr in Studio :( Aborting.")

        tgzr_exe = venv.get_exe("tgzr")
        tgzr_shortcut = self.path / f"tgzr_{project_name}"
        if tgzr_shortcut.exists():
            tgzr_shortcut.unlink()
        pm.create_shortcut(tgzr_exe, tgzr_shortcut, relative=True)

        install_errors = []
        all_ok = True
        for required_package in required_packages:
            print(f"--- Installing required package(s): {required_package}")
            ok = venv.install_packages(
                required_package,
                index=index,
                find_links=find_links,
                allow_prerelease=allow_prerelease,
                update=True,
            )
            if not ok:
                all_ok = False
                install_errors.append(required_package)
                print(
                    f"WARNING: Could not install {required_package!r} in Project {self.name!r} :("
                )
            else:
                print(f"Package(s) {required_package!r} installed successfully.")

        return all_ok, install_errors

    def get_venv_path(self, project_name: str | None) -> Path:
        if project_name is None:
            group = self._tgzr_venv_group
            venv_name = self._tgzr_venv_name
        else:
            group = self._projects_venv_group
            venv_name = project_name
        pm = PackageManager(self._venvs_path)
        return pm.get_venv_path(venv_name, group)

    def get_venv(self, project_name: str | None) -> Venv:
        """
        Return the package_manager.Venv for the given project.
        If project_name is not provided, the studio venv is used ('_/tgzr')
        """
        # if project_name is None:
        #     group = self._tgzr_venv_group
        #     venv_name = self._tgzr_venv_name
        # else:
        #     group = self._projects_venv_group
        #     venv_name = project_name
        # pm = PackageManager(self._venvs_path)
        # venv = pm.get_venv(venv_name, group)
        venv = Venv(self.get_venv_path(project_name=project_name))
        return venv

    def get_cmd_output(
        self, cmd_name: str, cmd_args: list[str], project_name: str | None = None
    ) -> tuple[str, str]:
        """
        Run a command in the venv for the given project and returns the stdout and stderr.
        If project_name is not provided, the studio venv is used ('_/tgzr')
        """
        venv = self.get_venv(project_name)
        return venv.get_cmd_output(cmd_name, cmd_args)

    def run_cmd(
        self, cmd_name: str, cmd_args: list[str], project_name: str | None = None
    ):
        """
        Run a command in the venv for the given project.
        If project_name is not provided, the studio venv is used ('_/tgzr')

        Returns True if the command was executed successfuly (according to process return int).
        """
        venv = self.get_venv(project_name)
        return venv.run_cmd(cmd_name, cmd_args)

    def get_packages(
        self, project_name: str | None = None
    ) -> list[importlib.metadata.Distribution]:
        """
        Return a list of (package_name, version, editable_path) for packages installed
        in the specified project.
        If project_name is not provided, the "master" venv is used ('_/tgzr')
        """
        venv = self.get_venv(project_name)
        return venv.get_packages()

    def get_plugins(
        self, project_name: str | None, group_filter: str | None
    ) -> list[tuple[importlib.metadata.EntryPoint, importlib.metadata.Distribution]]:
        venv = self.get_venv(project_name)
        return venv.get_plugins(group_filter)

    def shell(
        self,
        shell_type: Literal["xonsh", "bash", "cmd", "powershell", "auto"],
        project_name: str | None = None,
    ):
        """
        Open a shell in the venv for the given project.
        If project_name is not provided, the "master" venv is used ('_/tgzr')

        Returns True if the command was executed successfuly (according to process return int).
        """
        if project_name is None:
            group = self._tgzr_venv_group
            venv_name = self._tgzr_venv_name
        else:
            group = self._projects_venv_group
            venv_name = project_name

        pm = PackageManager(self._venvs_path)
        venv = pm.get_venv(venv_name, group)

        if shell_type == "auto":
            if platform.system() == "Windows":
                shell_type = "cmd"
            else:
                shell_type = "bash"

        if shell_type == "bash":
            activate = str(venv.get_exe("activate"))
            script = f"source /home/dee/.bashrc; source {activate}"
            try:
                fp, script_path = tempfile.mkstemp(prefix="tgzr_studio_shell_init_file")
                os.write(fp, script.encode())
            finally:
                os.close(fp)
            cmd = f"bash --init-file {script_path}"

        elif shell_type == "xonsh":
            xonsh = venv.get_exe("xonsh")
            if not xonsh.exists():
                venv.install_packages("xonsh[full]")
            cmd = f"{xonsh}"
        else:  # all windows shell types:
            activate = str(venv.get_exe("activate")).replace("/", "\\")
            cmd = f"start cmd /k {activate}"

        return venv.execute_cmd(cmd)



"""

Registers some generic tgzr.shell.app_plugin based on the current system:

"""

from typing import Callable
from types import ModuleType
import sys
import shutil
import warnings

from tgzr.shell.app_sdk.exe_app import ShellExeApp, DefaultShellAppInfo
from tgzr.shell.session import Session

from . import run_open, run_terminal


class OpenApp(ShellExeApp):
    def __init__(
        self,
        exe_path: str,
        color: str = "blue-grey-5",
        app_groups: set[str] = set(),
    ):
        app_groups.add("system")
        super().__init__(
            exe_path,
            app_name="open",
            run_module=run_open,
            app_groups=app_groups,
            default_app_info=DefaultShellAppInfo(
                icon="fa-regular fa-folder-open",
                color=color,
            ),
        )

    def cmd(self, session: Session, version: str | None) -> list[str]:
        # TODO: implement something like session.get_context_path(context)
        # and use it here.
        return super().cmd(session, version) + [str(session.home)]


class TerminalApp(ShellExeApp):
    def __init__(
        self,
        cmd_builder: Callable[[Session], list[str]],
        color: str = "blue-grey-5",
        app_groups: set[str] = set(),
    ):
        app_groups.add("system")
        super().__init__(
            exe_path=None,
            app_name="terminal",
            run_module=run_terminal,
            app_groups=app_groups,
            default_app_info=DefaultShellAppInfo(
                icon="fa-solid fa-terminal",
                color=color,
            ),
        )
        self._cmd_builder = cmd_builder

    def cmd(self, session: Session, version: str | None) -> list[str]:
        return self._cmd_builder(session)


def get_venv_activate_path(session: Session) -> str | None:
    studio = session.get_selected_studio()
    if studio is None:
        return None
    return str(studio.get_venv(session.selected_project_name).get_exe("activate"))


def linux_terminal_cmd(session: Session) -> list[str]:
    activate_script = get_venv_activate_path(session)
    activate_cmd = f"source {activate_script}; exec bash"

    terminal = shutil.which("x-terminal-emulator")
    if terminal:
        cmd = [terminal, "-e", "$SHELL", "-c", activate_cmd]
    else:
        known_terminals = [
            # GNOME (Ubuntu, Debian, Fedora default)
            # Modern gnome-terminal prefers '--' before the command arguments
            ("gnome-terminal", ["--", "bash", "-c", activate_cmd]),
            # KDE (Konsole)
            ("konsole", ["-e", "bash", "-c", activate_cmd]),
            # XFCE
            ("xfce4-terminal", ["-e", f"bash -c '{activate_cmd}'"]),
            # MATE
            ("mate-terminal", ["--", "bash", "-c", activate_cmd]),
            # Terminator
            ("terminator", ["-x", "bash", "-c", activate_cmd]),
            # xterm (Fallback)
            ("xterm", ["-e", f"bash -c '{activate_cmd}'"]),
        ]
        cmd = None
        for terminal, args in known_terminals:
            if shutil.which(terminal):
                cmd = [terminal] + args

    if cmd is None:
        raise SystemError("Could not find a terminal :/")

    print("---> terminal cmd:", cmd)
    return cmd


def win_terminal_cmd(session: Session) -> list[str]:
    activate_script = get_venv_activate_path(session)
    activate_cmd = f"cmd /K {activate_script}"
    return ["start", activate_cmd]


def mac_terminal_cmd(session: Session) -> list[str]:
    # We dont have a mac, let's hope this works.
    # Please fill an issue if it doesn't
    return linux_terminal_cmd(session)


apps = []
platform = sys.platform
if platform == "linux":
    color = "blue-10"
    open_app = OpenApp("xdg-open", color=color)
    terminal_app = TerminalApp(linux_terminal_cmd, color=color)
elif platform == "win32":
    color = "yellow-10"
    open_app = OpenApp("explorer.exe", color=color)
    terminal_app = TerminalApp(win_terminal_cmd, color=color)
elif platform == "darwin":
    color = "teal-10"
    open_app = OpenApp("open", color=color)
    terminal_app = TerminalApp(mac_terminal_cmd, color=color)
else:
    warnings.warn(
        f"Cannot register system apps: platform {platform!r} not supported yet."
    )


def register_apps():
    return [open_app, terminal_app]



from pathlib import Path

from tgzr.shell.app_sdk.nice_app import ShellNiceApp
from . import run_native, run_dev

from . import pages


app = ShellNiceApp(
    "shell_app_template",
    run_native_module=run_native,
    run_dev_module=run_dev,
    static_file_path=Path(pages.__file__).parent / "static_files",
)



from __future__ import annotations
from typing import TYPE_CHECKING, Type, Callable, Awaitable, Any

import os
import getpass
from pathlib import Path
import dataclasses
import logging

import pydantic

from .base_config import BaseConfig, Field, SettingsConfigDict
from .system import System
from .workspace import Workspace, WorkspaceConfig
from .studio import Studio, StudioConfig
from .project import Project, ProjectConfig
from .broker import AsyncBroker
from .services import ServiceClientPluginManager
from .settings import SettingsClientPlugin
from .connection import ConnectionPluginManager
from tgzr.package_management.plugin_manager import PluginManager

if TYPE_CHECKING:
    from .app import _BaseShellApp
    from .connection import Connection

logger = logging.getLogger(__name__)
_DEFAULT_SESSION: Session | None = None


class SessionConfig(BaseConfig):
    model_config = SettingsConfigDict(env_prefix="tgzr_")
    verbose: bool = Field(False, description="Tchatty logs")
    stream_name_prefix: str = "tgzr"
    subject_prefix: str = "tgzr.proto"


def set_default_session(session: Session):
    global _DEFAULT_SESSION
    _DEFAULT_SESSION = session
    logger.debug("Default TGZR Session set.")

    # Also set the env vars for subprocesses to build the session:
    os.environ.setdefault("tgzr_home", str(session.home))
    logger.info(f'  updated env "tgzr_home" to {str(session.home)}')

    if session.context.connection_url is not None:
        os.environ.setdefault(
            "tgzr_connection_url", str(session.context.connection_url)
        )
        logger.info(
            f'  updated env "tgzr_connection_url" to {str(session.context.connection_url)}'
        )
    if session.context.user_name is not None:
        os.environ.setdefault("tgzr_user_name", str(session.context.user_name))
        logger.info(
            f'  updated env "tgzr_user_name" to {str(session.context.user_name)}'
        )
    if session.context.studio_name is not None:
        os.environ.setdefault("tgzr_studio_name", str(session.context.studio_name))
        logger.info(
            f'  updated env "tgzr_studio_name" to {str(session.context.studio_name)}'
        )
    if session.context.project_name is not None:
        os.environ.setdefault("tgzr_project_name", str(session.context.project_name))
        logger.info(
            f'  updated env "tgzr_project_name" to {str(session.context.project_name)}'
        )


def get_default_session(ensure_set: bool = False) -> Session | None:
    """
    Return the session set as default with `set_default_session()`.
    If no session was set as default and the `tgzr_home` environment
    variable is set, a new session is created using it and set as
    default.
    If ensure_set is True, raise an EnvironmentError if no env var
    defines the session home.
    Return None in other cases.
    """
    if _DEFAULT_SESSION is None:
        env_home = os.environ.get("tgzr_home") or os.environ.get("TGZR_HOME")
        if env_home is None:
            if ensure_set:
                raise EnvironmentError(
                    "Missing 'TGZR_HOME' (or 'tgzr_home') env var. Cannot create a Session."
                )
            return None
        logger.info(
            f"Creating default TGZR session using 'tgzr_home' env var: {env_home}"
        )
        session = Session(home=env_home)
        set_default_session(session)

    return _DEFAULT_SESSION


class SessionContextConfig(BaseConfig):
    model_config = SettingsConfigDict(env_prefix="tgzr_")
    user_name: str = pydantic.Field(
        default_factory=getpass.getuser,
        description="The current use name (used in settings etc..)",
    )
    studio_name: str | None = pydantic.Field(
        default=None,
        description="The current studio name",
    )
    project_name: str | None = pydantic.Field(
        default=None,
        description="The current project name",
    )
    entity_name: str | None = pydantic.Field(
        default=None,
        description="The current entity name",
    )


@dataclasses.dataclass
class SessionContext:
    connection_url: str | None
    user_name: str
    studio_name: str | None = None
    project_name: str | None = None
    entity_name: str | None = None
    settings_base_context: tuple[str, ...] = ("system", "admin")


class Session:

    def __init__(self, home: Path | str | None = None):
        self._config_filename = ".tgzr"

        self._home: Path | None = None
        self._workspace: Workspace | None = None
        if home is None:
            home = Path.cwd().resolve()
        else:
            home = Path(home)
        self.set_home(home)

        self._known_config_types: list[Type[BaseConfig]] = []
        self.declare_config_type(SessionConfig)
        self.declare_config_type(SessionContextConfig)
        self.declare_config_type(WorkspaceConfig)
        self.declare_config_type(StudioConfig)
        self.declare_config_type(ProjectConfig)

        self._connection_plugin_manager = ConnectionPluginManager()

        context_config = SessionContextConfig()  # read env vars
        self.context = SessionContext(
            connection_url=None,
            user_name=context_config.user_name,
            studio_name=context_config.studio_name,
            project_name=context_config.project_name,
            entity_name=context_config.entity_name,
        )

        self._system: System = System(self)
        self._broker: AsyncBroker = AsyncBroker(self)
        self._service_plugin_manager = ServiceClientPluginManager()

        settings_plugins = self._service_plugin_manager.find_plugins(
            SettingsClientPlugin
        )
        if not settings_plugins:
            raise RuntimeError("Could not find a Settings plugin :/")
        if len(settings_plugins) > 1:
            raise RuntimeError(
                "Found more than one Settings plugin! (not yet supported)."
            )
        self._settings = settings_plugins[0]

        self._known_plugin_managers: list[PluginManager] = []
        self.declare_plugin_manager(self._broker._plugin_manager)
        self.declare_plugin_manager(self._service_plugin_manager)
        # TODO: refactor cli plugins to use PluginManager and declare it:
        # self.declare_plugin_manager(the_cli_plugin_manager)
        # TODO: refactor apps plugins to use PluginManager and declare it:
        # self.declare_plugin_manager(self._apps_plugin_manager)

    #
    # FILESYSTEM
    #

    @property
    def home(self) -> Path:
        if self._home is None:
            raise ValueError("Home not set. Please use `set_home(path) first.`")
        return self._home

    @property
    def system(self) -> System:
        return self._system

    @property
    def workspace(self) -> Workspace:
        if self._workspace is None:
            raise RuntimeError("Session workspace not yet configured.")
        return self._workspace

    def set_home(
        self, search_path: str | Path, ensure_config_found: bool = True
    ) -> Path | None:
        """Returns the path of the config loaded or None."""
        # FIXME: `ensure_config_found` arg is probably a bad idea / not needed anymore

        search_path = Path(search_path).resolve()
        config_path = SessionConfig.find_config_file(search_path, self._config_filename)
        if config_path is None:
            if ensure_config_found:
                raise FileNotFoundError(
                    f"Could not find a {self._config_filename} file under {search_path!r} and ancestors. "
                    "(Use `ensure_config_found=False` to bypass this error.)"
                )
            self._home = search_path
        else:
            self._home = config_path.parent
            self._config = SessionConfig.from_file(config_path)

        try:
            self._workspace = Workspace(self)
        except FileNotFoundError:
            self._workspace = None

        # If the given path is inside a studio, let's make this studio
        # the default one:
        # FIXME: use self.context here + work out precedence between env, cli args, current dir, home, etc... + do it for project too!
        if self._workspace and search_path.is_relative_to(self._workspace.path):
            studio_name = str(search_path)[len(str(self._workspace.path)) + 1 :].split(
                os.path.sep, 1
            )[0]
            try:
                self._workspace.set_default_studio(
                    studio_name,
                    ensure_exists=True,
                    save_config=False,  # Do not save, this is just for this session.
                )
            except FileNotFoundError:
                pass
        return config_path

    #
    # CONFIG
    #

    @property
    def config(self) -> SessionConfig:
        if self._config is None:
            raise ValueError("Config not set. Please use `set_home(path) first.`")
        return self._config

    def save_config(self) -> Path:
        """Returns the path of the saved file."""
        return self.write_config_file(None, allow_overwrite=True)

    def write_config_file(
        self, path: str | Path | None, allow_overwrite: bool = False
    ) -> Path:
        """Returns the path of the saved file."""
        if path is None:
            path = self.home / self._config_filename
        path = Path(path).resolve()
        self._config.to_file(
            path,
            allow_overwrite=allow_overwrite,
            header_text="tgzr session config",
        )
        return path

    def declare_config_type(self, config_type: Type[BaseConfig]):
        """
        Declare the config type as used in the session.
        This is used for documentation and cli help.
        """
        self._known_config_types.append(config_type)

    def get_config_env_vars(self) -> list[tuple[str, list[tuple[str, str | None]]]]:
        """
        Returns a list of (name, description) for each config
        declared with `declare_config_type()`
        """
        # TODO: use config type's docstring to show their description in `tgzr help env`
        config_env_vars = []
        for config_type in self._known_config_types:
            config_env_vars.append((config_type.__name__, config_type.used_env_vars()))
        return config_env_vars

    def declare_plugin_manager(self, plugin_manager: PluginManager):
        # TODO: add something like self.get_plugin_manager_info and use it in cli/manager_panel
        self._known_plugin_managers.append(plugin_manager)

    def get_plugin_managers(self) -> list[PluginManager]:
        return self._known_plugin_managers.copy()

    def set_verbose(self):
        self.config.verbose = True

    def set_quiet(self):
        self.config.verbose = False

    @property
    def quiet(self) -> bool:
        return not self.config.verbose

    @property
    def verbose(self) -> bool:
        return self.config.verbose

    #
    # CONTEXT
    #

    def select_studio(self, studio_name: str | None = None):
        self.context.studio_name = studio_name

    @property
    def selected_studio_name(self) -> str | None:
        return self.context.studio_name

    def get_selected_studio(self, ensure_exists: bool = True) -> Studio | None:
        if self.workspace is None:
            return None
        studio_name = self.context.studio_name
        if studio_name is None:
            studio_name = self.workspace.default_studio_name()
        if studio_name is None:
            return None
        return self.workspace.get_studio(studio_name, ensure_exists)

    @property
    def selected_project_name(self) -> str | None:
        return self.context.project_name

    def select_project(self, project_name: str | None = None):
        self.context.project_name = project_name

    def get_selected_project(self) -> Project | None:
        if self.workspace is None:
            return None

        if self.context.project_name is None:
            return None

        studio = self.get_selected_studio()
        if studio is None:
            return None

        return studio.get_project(self.context.project_name)

    #
    #   BROKER
    #

    async def connect(self, url: str | None = None):
        """
        Connect all the services.

        Auth is first done on url, and broket configuration
        is fetched from there. Then settings connects and all
        other services/clients use settings to get they configuration.

        If url is None, session.context.connection_url is used.
        How that url is used depends on the list of installed ConnectionPlugin
        (see tgzr.shell.connection.ConnectionPlugin).

        The session.context.connection_url set to url before connection
        attempt.
        """
        if self._broker.connected():
            logger.warning(
                "!!!! Sesssion already connected !!!\n(Please avoid creating duplicate session/connection)"
            )
            return

        if url is None:
            url = self.context.connection_url
        self.context.connection_url = url

        logger.info(f"Connecting session to {self.context.connection_url}")
        connection = self._connection_plugin_manager.get_connection(
            self, self.context.connection_url
        )
        logger.info("Fetching Broker config...")
        broker_connection_config = await connection.get_broker_connection_config()
        logger.info("Connecting Broker...")
        await self._broker.connect(broker_connection_config)
        logger.info("Connecting Settings client...")
        await self.settings.connect(self)
        logger.info("Connecting all Services...")
        for service_client in self._service_plugin_manager.get_plugins():
            if service_client is self.settings:
                continue
            await service_client.connect(self)

    async def disconnect(self):
        """Disconnect all the services."""
        logger.info("Closing all services.")
        for service_client in self._service_plugin_manager.get_plugins():
            await service_client.disconnect()
        await self._broker.disconnect()

    async def subscribe(
        self,
        subject_pattern: str,
        callback: Callable[[AsyncBroker.Event], Awaitable[None]],
    ) -> AsyncBroker.Subscription:
        return await self._broker.subscribe(
            subject_pattern=subject_pattern, callback=callback
        )

    async def publish(self, subject: str, **data) -> None:
        await self._broker.publish(subject=subject, **data)

    async def unsubscribe(self, subscription: AsyncBroker.Subscription) -> None:
        await self._broker.unsubscribe(subscription=subscription)

    async def execute_service_cmd(self, service_name, cmd_name, **cmd_kwargs) -> None:
        await self._broker.cmd(
            service_name=service_name, cmd_name=cmd_name, **cmd_kwargs
        )

    async def execute_service_query(
        self, service_name, query_name, **query_kwargs
    ) -> Any:
        data = await self._broker.query(
            service_name=service_name, query_name=query_name, **query_kwargs
        )
        return data

    #
    # MANAGED SERVICES
    #

    @property
    def settings(self) -> SettingsClientPlugin:
        return self._settings

    #
    # APPS
    #

    def apps(self) -> list[_BaseShellApp]:
        from .app import get_apps

        return get_apps()



from __future__ import annotations
from typing import TYPE_CHECKING, Any, Type, TypeVar

import importlib.metadata
import warnings

import pydantic

from ..broker import AsyncBroker

if TYPE_CHECKING:
    from ..session import Session
    from .settings_provider import BaseSettingsProvider
    from .settings_plugin import SettingsProviderPlugin

SettingsModelType = TypeVar("SettingsModelType", bound=pydantic.BaseModel)

# TODO: implement a settings getter like this ?:
# getter = SettingsGetter(context, path)
# anim_settings = getter / project / seq / shot/ 'Tasks'/'Anim'
# framerate = anim_settings.get('frame_rate')
# render_param:MyRenderSettingsModel = anim_settings.get_model('render')
# ---> this is not interesting when using Models for settings
# ------> but an intermediate class like this could be nice:
# my_settings = SettingsFor(context, path, ModelType)
# my_settings.data.frame_rate = 12
# my_settings.update(context_name)
# my_settings.merge(from_context_name, in_context_name)
# my_settings.duplicate(context_name, as_context_name)

# TODO:
# We should probably store app settings and other settings in separate stores.
# So we could have
# - `Session.system_settings`: connection settings, event broken, (plugin selection config)?, etc...
# - `Session.shell_app_settings`: all shell app settings
# - `Session.requirement`: index&index auth + per studio/project requirements + auto-sync packages settings
# etc...


# class Settings:
#     _PROVIDER_PLUGINS: list[SettingsProviderPlugin] | None = None

#     @classmethod
#     def load_settings_provider_plugins(cls, session: Session):
#         entry_point_group = "tgzr.shell.settings_provider"
#         # print(f"Loading Settings Provider plugins from {entry_point_group}")

#         all_entry_points = importlib.metadata.entry_points(group=entry_point_group)

#         plugins = []
#         errs = []
#         for ep in all_entry_points:
#             # print(f"   -> EP: {ep}")
#             try:
#                 plugin_getter = ep.load()
#             except Exception as err:
#                 errs.append((ep, err))
#                 warnings.warn(f"Error loading plugin {ep}: {err}")
#             else:
#                 plugins.append(plugin_getter(session))
#         cls._PROVIDER_PLUGINS = plugins
#         cls._BROKEN_PROVIDER_PLUGINS = errs

#     @classmethod
#     def get_settings_provider_plugin(
#         cls, session: Session, url: str, refresh_plugin_list: bool = False
#     ) -> BaseSettingsProvider:
#         if cls._PROVIDER_PLUGINS is None or refresh_plugin_list:
#             cls.load_settings_provider_plugins(session)
#         if cls._PROVIDER_PLUGINS is None:
#             raise Exception("No settings providers found.")

#         # print(f"Looking for settings provider plugin in {cls._PROVIDER_PLUGINS}")
#         for provider_plugin in cls._PROVIDER_PLUGINS:
#             # print(f"Settings for {url}?", provider_plugin)
#             if provider_plugin.handles(url):
#                 # print("  YES!")
#                 provider = provider_plugin.create_provider(url)
#                 return provider
#             # print("  nope.")
#         raise Exception(f"No settings providers to handle settings with {url=}.")

#     def __init__(self, session: Session):
#         self._session = session

#         settings_url = ":memory:"  # TODO: get it from session.config
#         self._provider = self.get_settings_provider_plugin(session, settings_url)

#         self._cmd_sub: AsyncBroker.Subscription = None

#     async def connect(self):
#         self._cmd_sub = await self._session.subscribe(
#             "$CMD.session.settings.*", self.on_cmd
#         )

#     async def close(self):
#         if self._cmd_sub is not None:
#             await self._session.unsubscribe(self._cmd_sub)

#     def info(self) -> dict[str, Any]:
#         return dict(
#             provider=self._provider,
#             url=self._provider.url,
#         )

#     def get_context_names(self) -> tuple[str, ...]:
#         return self._provider.get_context_names()

#     def set_context_info(self, context_name: str, **info: Any) -> None:
#         return self._provider.set_context_info(context_name, **info)

#     def get_context_info(self, context_name: str) -> dict[str, Any]:
#         return self._provider.get_context_info(context_name)

#     def expand_context_name(self, context_name: str) -> list[str]:
#         return self._provider.expand_context_name(context_name)

#     # -- OPERATIONS

#     def set(self, context_name: str, key: str, value: Any) -> None:
#         self._provider.set(context_name, key, value)

#     def toggle(self, context_name: str, key: str) -> None:
#         self._provider.toggle(context_name, key)

#     def add(self, context_name: str, key: str, value: Any) -> None:
#         self._provider.add(context_name, key, value)

#     def sub(self, context_name: str, key: str, value: Any) -> None:
#         self._provider.sub(context_name, key, value)

#     def set_item(
#         self, context_name: str, key: str, index: int, item_value: Any
#     ) -> None:
#         self._provider.set_item(context_name, key, index, item_value)

#     def del_item(self, context_name: str, key: str, index: int) -> None:
#         self._provider.del_item(context_name, key, index)

#     def remove(self, context_name: str, key: str, item: str) -> None:
#         self._provider.remove(context_name, key, item)

#     def append(self, context_name: str, key: str, value: Any) -> None:
#         self._provider.append(context_name, key, value)

#     def env_override(self, context_name: str, key: str, envvar_name: str) -> None:
#         self._provider.env_override(context_name, key, envvar_name)

#     def pop(self, context_name: str, key: str, index: int | slice) -> None:
#         self._provider.pop(context_name, key, index)

#     def remove_slice(
#         self,
#         context_name: str,
#         key: str,
#         start: int,
#         stop: int,
#         step: int | None = None,
#     ) -> None:
#         self._provider.remove_slice(context_name, key, start, stop, step)

#     def call(
#         self,
#         context_name: str,
#         key: str,
#         method_name: str,
#         args: list[Any],
#         kwargs: dict[str, Any],
#     ) -> None:
#         self._provider.call(context_name, key, method_name, args, kwargs)

#     def update_flat(
#         self, context_name: str, flat_dict: dict[str, Any], path: str | None = None
#     ) -> None:
#         self._provider.update_flat(context_name, flat_dict, path)

#     def update_deep(
#         self,
#         context_name: str,
#         deep_dict: dict[str, Any | dict[str, Any]],
#         path: str | None = None,
#     ) -> None:
#         self._provider.update_deep(context_name, deep_dict, path)

#     def update(
#         self,
#         context_name: str,
#         model: pydantic.BaseModel,
#         path: str | None = None,
#         exclude_defaults: bool = True,
#     ):
#         """
#         Update the given context with value from the given model.

#         If `exclude_unset` is True, only non-default values will
#         be stored.
#         If you need to store a default value without storing all
#         the default values, dump the model yourself keeping only
#         the fields you want to store, and use `update_context_dict()`.
#         """
#         self._provider.update(context_name, model, path, exclude_defaults)

#     def get_flat(
#         self,
#         context: list[str],
#         key: str | None = None,
#         with_history: bool = False,
#     ) -> dict[str, Any]:
#         return self._provider.get_flat(context, key, with_history)

#     def get_deep(
#         self,
#         context: list[str],
#         key: str | None = None,
#         with_history: bool = False,
#     ) -> dict[str, Any]:
#         return self._provider.get_deep(context, key, with_history)

#     def get_context(
#         self,
#         context: list[str],
#         model_type: type[SettingsModelType],
#         key: str | None = None,
#     ) -> SettingsModelType:
#         """
#         Return the value for the given context.

#         The `context` argument is a list of context name to apply.
#         Each name can contain envvars and/or path to expand.
#         See `context.expand_context_names()` doc for details.

#         If `key` is not None, it is the dotted name of
#         a value in the store.
#         If `key` is None, the a value containing all values is returned.

#         If `model_type` is provided, it must be a pydantic.BaseModel and
#         the returned value will be an instance of it.
#         This model *must* be have default for *all fields*
#         (i.e `model_type()` must be valid)

#         Note that the return value is validated against model_type and a
#         pydantic.ValidationError may be raised.
#         If you need to access value without validation, you must
#         use `get_context_dict()` instead.
#         """
#         return self._provider.get_context(context, model_type, key)

#     async def on_cmd(self, event: AsyncBroker.Event):
#         """
#         Called when a cmd event has been published to $CMD.session.settings
#         """
#         op, context_name, key, kwargs = event.unpack("op", "context_name", "key")
#         meth = getattr(self, op)
#         meth(context_name=context_name, key=key, **kwargs)
#         await self._session.publish(
#             "session.settings.changed", context_name=context_name, key=key
#         )


from ..services import ServiceClientPlugin


ModelType = TypeVar("ModelType", bound=pydantic.BaseModel)


class SettingsClientPlugin(ServiceClientPlugin):
    @classmethod
    def plugin_type_name(cls) -> str:
        return "ServiceClientPlugin"

    async def get_context_names(self) -> tuple[str, ...]: ...
    async def set_context_info(self, context_name: str, **info: Any) -> None: ...
    async def get_context_info(self, context_name: str) -> dict[str, Any]: ...
    def expand_context_name(self, context_name: str) -> list[str]: ...

    #
    # ---
    #
    async def get_context_flat(
        self,
        context: list[str],
        path: str | None = None,
        with_history: bool = False,
    ) -> dict[str, Any]: ...

    async def get_context_dict(
        self, context: list[str], path: str | None = None, with_history: bool = False
    ) -> dict[str, Any]: ...

    async def get_context(
        self,
        context: list[str],
        model_type: type[ModelType],
        path: str | None = None,
    ) -> ModelType: ...

    #
    # ---
    #

    async def set(self, context_name: str, name: str, value: Any) -> None: ...
    async def toggle(self, context_name: str, name: str) -> None: ...
    async def add(self, context_name: str, name: str, value: Any) -> None: ...
    async def sub(self, context_name: str, name: str, value: Any) -> None: ...
    async def set_item(
        self, context_name: str, name: str, index: int, item_value: Any
    ) -> None: ...
    async def del_item(self, context_name: str, name: str, index: int) -> None: ...
    async def remove(self, context_name: str, name: str, item: str) -> None: ...
    async def append(self, context_name: str, name: str, value: Any) -> None: ...
    async def env_override(
        self, context_name: str, name: str, envvar_name: str
    ) -> None: ...
    async def pop(self, context_name: str, name: str, index: int | slice) -> None: ...
    async def remove_slice(
        self,
        context_name: str,
        name: str,
        start: int,
        stop: int,
        step: int | None = None,
    ) -> None: ...
    async def call(
        self,
        context_name: str,
        name: str,
        method_name: str,
        args: list[Any],
        kwargs: dict[str, Any],
    ) -> None: ...


# from tgzr.contextual_settings.stores.jetstream_store import (
#     JetStreamStoreClient,
#     ClientBroker,
# )

# class _TgzrSessionSettingsBroker(ClientBroker):
#     """
#     This is used to connect the JetStreamStore to the session's broker.

#     See `tgzr.contextual_settings.stores.jetstream_store.ClientBroker()`
#     """

#     def __init__(self, stream_name: str, subject_prefix: str):
#         super().__init__(stream_name, subject_prefix)
#         self._session: Session = None  # type: ignore

#         self._service_name = subject_prefix
#         self._cmd_subject_prefix = f"{subject_prefix}.$CMD."
#         self._request_subject_prefix = f"{subject_prefix}.$QUERY."
#         self._touch_event_prefix = f"{subject_prefix}.$EVENT."

#     async def connect(self, session: Session) -> None:
#         self._session = session
#         self._touch_subscription = await session.subscribe(
#             self._touch_event_prefix + ">", self._on_touch_event
#         )

#     async def disconnect(self) -> None:
#         if self._touch_subscription is not None:
#             await self._session.unsubscribe(self._touch_subscription)

#     async def send_cmd(self, cmd_name, **cmd_kwargs) -> None:
#         await self._session.execute_service_cmd(
#             self._service_name, cmd_name, **cmd_kwargs
#         )

#     async def send_query(self, query_name: str, **query_kwargs) -> Any:
#         data = await self._session.execute_service_query(
#             self._service_name, query_name, **query_kwargs
#         )
#         return data

#     async def _on_touch_event(self, msg) -> None:
#         print("SETTINGS TOUCH EVENT:", msg)


# class SettingsClient(ServiceClientPlugin):
#     def __init__(self, session: Session):
#         self._session = session

#         # TODO: use plugin to select the store?
#         self._store: JetStreamStoreClient = None  # type: ignore

#     async def connect(self):
#         # TODO: use a plugin to select / configure the store?
#         session_broker = _TgzrSessionSettingsBroker(
#             stream_name="test_settings", subject_prefix="test.settings.proto"
#         )
#         self._store = JetStreamStoreClient(session_broker)
#         await self._store.connect(self._session)

#     async def disconnect(self):
#         if self._store is not None:
#             await self._store.disconnect()

#     def info(self) -> dict[str, Any]:
#         return dict(
#             provider=self._store.__class__.__name__,
#             url="*******",
#         )

#     @property
#     def _(self) -> JetStreamStoreClient:
#         if self._store is None:
#             raise RuntimeError(
#                 "SettingsClient not connected. Please call connect() first!"
#             )
#         return self._store

#     #
#     # ---
#     #

#     async def get_context_names(self) -> tuple[str, ...]:
#         return await self._.get_context_names()

#     async def set_context_info(self, context_name: str, **info: Any) -> None:
#         return await self._.set_context_info(context_name, **info)

#     async def get_context_info(self, context_name: str) -> dict[str, Any]:
#         return await self._.get_context_info(context_name)

#     def expand_context_name(self, context_name: str) -> list[str]:
#         return self._.expand_context_name(context_name)

#     #
#     # ---
#     #
#     async def get_context_flat(
#         self,
#         context: list[str],
#         path: str | None = None,
#         with_history: bool = False,
#     ) -> dict[str, Any]: ...

#     async def get_context_dict(
#         self, context: list[str], path: str | None = None, with_history: bool = False
#     ) -> dict[str, Any]: ...

#     async def get_context(
#         self,
#         context: list[str],
#         model_type: type[ModelType],
#         path: str | None = None,
#     ) -> ModelType: ...

#     #
#     # ---
#     #

#     async def set(self, context_name: str, name: str, value: Any) -> None: ...
#     async def toggle(self, context_name: str, name: str) -> None: ...
#     async def add(self, context_name: str, name: str, value: Any) -> None: ...
#     async def sub(self, context_name: str, name: str, value: Any) -> None: ...
#     async def set_item(
#         self, context_name: str, name: str, index: int, item_value: Any
#     ) -> None: ...
#     async def del_item(self, context_name: str, name: str, index: int) -> None: ...
#     async def remove(self, context_name: str, name: str, item: str) -> None: ...
#     async def append(self, context_name: str, name: str, value: Any) -> None: ...
#     async def env_override(
#         self, context_name: str, name: str, envvar_name: str
#     ) -> None: ...
#     async def pop(self, context_name: str, name: str, index: int | slice) -> None: ...
#     async def remove_slice(
#         self,
#         context_name: str,
#         name: str,
#         start: int,
#         stop: int,
#         step: int | None = None,
#     ) -> None: ...
#     async def call(
#         self,
#         context_name: str,
#         name: str,
#         method_name: str,
#         args: list[Any],
#         kwargs: dict[str, Any],
#     ) -> None: ...


async def test():
    from ..session import Session

    session = Session()
    print(session.settings)
    await session.connect()
    print("Context names:", await session.settings.get_context_names())
    # if 1:
    #     print(await session.settings.set_context_info("test_context", color="red"))
    print("Context info:", await session.settings.get_context_info("test_context"))

    await session.disconnect()


if __name__ == "__main__":
    import asyncio

    asyncio.run(test())



from __future__ import annotations
from typing import TYPE_CHECKING, Any
import logging

from ..connection import ConnectionPlugin, Connection

if TYPE_CHECKING:
    from ..session import Session

logger = logging.getLogger(__name__)


class DefaultConnectionPlugin(ConnectionPlugin):
    """
    This plugin is used when no connection url is provided
    (url is None).
    """

    @classmethod
    def handles(cls, url_scheme: str | None) -> bool:
        return url_scheme is None

    def get_connection(self, session: Session, url: str | None) -> DefaultConnection:
        return DefaultConnection(session, url)


class DefaultConnection(Connection):

    def __init__(self, session: Session, url: str | None):
        super().__init__(session, url="*default*")

    async def get_broker_connection_config(self) -> dict[str, Any]:
        broker_plugin_manager = self.session._broker._plugin_manager
        broker = self.session._broker._broker_implementation
        broker_config_path = self.session.system.get_plugin_config_path(
            broker_plugin_manager,
            broker,
        )
        logger.info(f"Reading broker config from {broker_config_path}")
        user_credentials_file = broker_config_path / "broker.creds"
        if not user_credentials_file.exists():
            raise ValueError(
                f"Cannot connect broker, missing credential file: {user_credentials_file}"
            )

        # TODO: connect to tgzr.org instead of the broker, then get the creds from there!
        config = dict(
            servers="tls://connect.ngs.global",
            user_credentials=str(user_credentials_file),
        )
        return config



if __name__ == "__main__":
    from .register import open_app
    from tgzr.shell.session import get_default_session

    session = get_default_session(ensure_set=True)
    open_app.run_app(session=session, version=None)



from pathlib import Path

import click
import rich

from tgzr.cli.utils import TGZRCliGroup

from .utils import pass_session, Session


@click.group(cls=TGZRCliGroup, help="Manage tgzr session configs.")
def session_group():
    pass


@session_group.command()
@pass_session
def show(session):
    rich.print(session.config)


@session_group.command()
@click.option("-o", "--output", help='The file to save. Defaults to "home/.tgzr".')
@click.option("--force", is_flag=True)
@pass_session
def save(session: Session, output: Path | None = None, force: bool = False):
    try:
        session.write_config_file(output, allow_overwrite=force)
    except FileExistsError as err:
        click.echo(f"File {err} already exists! (Use --force to allow overwrite.)")



from typing import TYPE_CHECKING

import click
import rich
import rich.table, rich.tree, rich.pretty
import importlib.metadata

from tgzr.cli.utils import TGZRCliGroup

from .utils import pass_session, Session

if TYPE_CHECKING:
    from tgzr.package_management.plugin_manager import PluginManager, Plugin


def dict_table(d):
    t = rich.table.Table.grid(
        rich.table.Column("key", justify="right"), "value", padding=(0, 1)
    )
    for k, v in d.items():
        t.add_row(f"[cyan]{k}:", str(v))
    return t


@click.group(cls=TGZRCliGroup, help="Manage plugins")
def plugins_group():
    pass


@plugins_group.command()
@pass_session
def ls(session: Session):
    """
    List information about all plugin managers.
    """
    table = rich.table.Table("Name", "Entry Point", "Installed", "Broken")
    for plugin_manager in session.get_plugin_managers():
        broken = plugin_manager.get_broken_plugins()
        nb_broken = len(broken)
        if nb_broken == 1:
            nb_broken = str(broken[0])

        plugins = plugin_manager.get_plugins()
        nb_plugins = len(plugins)
        if nb_plugins == 1:
            nb_plugins = str(plugins[0])

        table.add_row(
            plugin_manager.managed_plugin_type().plugin_type_name(),
            plugin_manager.EP_GROUP,
            str(nb_plugins),
            str(nb_broken),
        )
    rich.print(table)


@plugins_group.command()
@click.argument("filter", required=False)
@pass_session
def show(session: Session, filter: str | None = None):
    """
    Show detailed information about plugin manager.
    Use [FILTER] to filter out based on entry points.
    """
    for plugin_manager in session.get_plugin_managers():
        if filter is not None and filter not in plugin_manager.EP_GROUP:
            continue

        manager_name = plugin_manager.managed_plugin_type().plugin_type_name()

        plugins: list[Plugin] = plugin_manager.get_plugins()
        if plugins:
            loaded_table = rich.table.Table(
                "Loaded",
                title=f"{plugin_manager.EP_GROUP} (loaded by [b]{manager_name}[/b])",
            )
            for plugin in plugins:
                info = plugin.plugin_info()
                info["instance"] = plugin
                info["package"] = plugin.__module__
                try:
                    info["dist-version"] = importlib.metadata.distribution(
                        plugin.__module__
                    ).version
                except:
                    info["dist-version"] = "no that easy to find..."
                loaded_table.add_row(
                    dict_table(info),
                )
            rich.print(loaded_table)

        broken = plugin_manager.get_broken_plugins()
        if broken:
            broken_table = rich.table.Table("Name", title=f"Broken in {manager_name}")
            for ep, exception in broken:
                broken_table.add_row(str(exception))

            rich.print(broken_table)



import click
import rich
import rich.table


from .utils import pass_session, Session


@click.command("env")
@pass_session
def env_help(session: Session):
    config_env_vars = session.get_config_env_vars()

    for name, env_vars in config_env_vars:
        table = rich.table.Table("Name", "Description", title=name)

        for name, description in env_vars:
            table.add_row(name, description or "")

        rich.print("\n", table)



from typing import cast

from types import ModuleType
from ._base_app import _BaseShellApp, DefaultShellAppInfo


class ShellQtpyApp(_BaseShellApp):
    def __init__(
        self,
        app_name: str,
        run_module: ModuleType | None,
        app_id: str | None = None,
        app_groups: set[str] = set(),
        default_app_info: DefaultShellAppInfo | None = None,
    ):
        app_groups.add("_QT_")
        super().__init__(app_name, run_module, app_id, app_groups, default_app_info)

    def run_app(self, main_widget: object):
        # NOTE:
        # The nice gui app must be runnable with:
        # `python -m <self.run_native_module>`

        # Do the import here instead of global to have the module
        # loadable even without the [nicegui] extra requirements.
        # (I know...)
        from qtpy.QtWidgets import QWidget

        widget = cast(QWidget, main_widget)
        raise NotImplementedError()
        # TODO: implement this:
        # Add a 'qtpy' extra requirement with qtpy
        # Build a QApplication here, instantiate self.main_widget
        # Run the app.



from typing import Type, Callable, Iterable
import importlib.metadata

from .app_sdk._base_app import _BaseShellApp
from .app_sdk.nice_app import ShellNiceApp
from .app_sdk.qtpy_app import ShellQtpyApp
from .app_sdk.exe_app import ShellExeApp
from .app_sdk.host_app import ShellHostApp

_BROKEN_APPS: list[tuple[importlib.metadata.EntryPoint, Exception]] | None = None
_LOADED_APPS: list[_BaseShellApp] | None = None


def _get_plugin_apps(
    loaded: (
        _BaseShellApp
        | Callable[[], _BaseShellApp | list[_BaseShellApp]]
        | Iterable[_BaseShellApp]
    ),
) -> list[_BaseShellApp]:
    # print("Resolving shell app plugins:", loaded)
    if isinstance(loaded, _BaseShellApp):
        # print("  is app")
        return [loaded]

    if callable(loaded):
        # print("  is callable")
        return _get_plugin_apps(loaded())

    if isinstance(loaded, (tuple, list, set)):
        # print("  is iterable")
        return [app for app in loaded]

    # print("  is unsupported")
    raise ValueError(
        'Invalid value for "tgzr.shell.app_plugin" entry point. '
        f"Must be a {_BaseShellApp}, a list/tuple/set of {_BaseShellApp}, or a callable returning one of these"
    )


def _load_apps_plugins():
    global _BROKEN_APPS, _LOADED_APPS

    entry_point_group = "tgzr.shell.app_plugin"

    all_entry_points = importlib.metadata.entry_points(group=entry_point_group)

    apps = []
    errs = []
    for ep in all_entry_points:
        # print(f"Loading {entry_point_group} plugin:", ep.name)
        try:
            loaded = ep.load()
        except Exception as err:
            errs.append((ep, err))
        else:
            try:
                plugin_apps = _get_plugin_apps(loaded)
            except Exception as err:
                errs.append((ep, err))
            for app in plugin_apps:
                apps.append(app)
    _LOADED_APPS = apps
    _BROKEN_APPS = errs


def get_broken_apps() -> list[tuple[importlib.metadata.EntryPoint, Exception]]:
    if _BROKEN_APPS is None:
        _load_apps_plugins()
    return _BROKEN_APPS  # type: ignore


def get_apps() -> list[_BaseShellApp]:
    if _LOADED_APPS is None:
        _load_apps_plugins()
    return _LOADED_APPS  # type: ignore



import os
import platform
from pathlib import Path
import subprocess
import importlib.metadata
import json

import uv


class Venv:
    def __init__(self, path: Path | str) -> None:
        self._path = Path(path)
        if platform.system() == "Windows":
            self._bin_path = self._path / "Scripts"
        else:
            self._bin_path = self._path / "bin"
        self._site_packages_path: Path | None = None

    @property
    def path(self) -> Path:
        return self._path

    @property
    def site_packages_path(self) -> Path | None:
        if self._site_packages_path is None:
            # Note:
            # On Windows: venv/Lib/site-packages
            # On POSIX:   venv/lib/pythonX.Y/site-packages
            site_packages_path = self._path / "Lib" / "site-packages"
            if site_packages_path.exists():
                self._site_packages_path = site_packages_path
            else:
                lib = self._path / "lib"
                for p in lib.glob("python*"):
                    site_packages_path = p / "site-packages"
                    if site_packages_path.is_dir():
                        self._site_packages_path = site_packages_path
                        break
        return self._site_packages_path

    def exists(self) -> bool:
        return (
            self.path.exists()
            and self._bin_path.exists()
            and self.site_packages_path is not None
        )

    def create(self, prompt: str | None = None, clear_existing: bool = False):
        uv_exe = uv.find_uv_bin()

        prompt_option = ""
        if prompt:
            prompt_option = f"--prompt {prompt}"

        clear_option = ""
        if clear_existing:
            clear_option = f"--clear"

        cmd = f"{uv_exe} venv --seed {clear_option} {prompt_option} {self.path}"
        print(f"Creating venv {self.path}: {cmd}")
        os.system(cmd)

    def get_exe(self, name: str):
        exe = self._bin_path / name
        if platform.system() == "Windows":
            # the '.exe' suffix is not needed to execute
            # but the caller may want to use `exe.exists()``
            # so we need to add the extension if on windows :/
            exe_suffix = exe.with_suffix(".exe")
            if exe_suffix.exists():
                # special case for .bat files
                exe = exe_suffix
        return exe

    def install_uv(self):
        return self.install_packages("uv", use_uv=False)

    def install_packages(
        self,
        requirements,
        update: bool = True,
        use_uv: bool = True,
        index: str | None = None,
        find_links: str | None = None,
        allow_prerelease: bool = False,
    ) -> bool:
        """
        Returns True if the package has been successfully installed.

        if `index` is given, it will be passed to pip:
            - as -f/--find-links it it is a path (Note: relative path are discouraged).
            - as -i/--index-url otherwise
        """
        if use_uv:
            # NB: we're using python -m uv instead of uv directly so that uv pip installs in the same env
            # see "If uv is installed in a Python environment" in https://docs.astral.sh/uv/pip/environments/#using-arbitrary-python-environments
            exe = f'{self.get_exe("python")} -m uv pip'
        else:
            exe = f'{self.get_exe("python")} -m pip'

        index_flags = ""
        if index is not None:
            if "://" in index:
                index_flags = f"--default-index {index}"
            else:
                index_flags = f"--find-links {index}"

        find_links_options = ""
        if find_links:
            find_links_options = f"--find-links {find_links}"

        prerelease_options = ""
        if allow_prerelease:
            prerelease_options = "--prerelease=allow"

        update_flag = ""
        if update:
            update_flag = "-U"

        cmd = f"{exe} install {index_flags} {update_flag} {find_links_options} {prerelease_options} {requirements}"
        print(f"Installing package(s) {requirements}: {cmd}")
        ret = os.system(cmd)
        return not ret

    def execute_cmd(self, cmd: str) -> bool:
        """
        Returns True if the command was been successfully executed.
        """
        print("Executing venv cmd:", cmd)
        ret = os.system(cmd)
        return not ret

    def get_cmd_output(self, cmd_name: str, cmd_args: list[str]) -> tuple[str, str]:
        """
        Returns the stdout and stderr of a the command.
        """
        # TODO: thread this.
        exe = str(self.get_exe(cmd_name))
        # subprocess.Popen(text=True)
        result = subprocess.run(
            [exe] + cmd_args, capture_output=True, check=True, text=True
        )
        print("CMD", " ".join([exe] + cmd_args), "->", result.returncode)
        return result.stdout, result.stderr

    def run_cmd(self, cmd_name: str, cmd_args: list[str]) -> bool:
        """
        Returns True if the command was been successfully executed.
        """
        # TODO: thread this.
        exe = self.get_exe(cmd_name)
        cmd = str(exe) + " " + " ".join(cmd_args)
        return self.execute_cmd(cmd)

    def get_cmd_names(self) -> list[str]:
        raise NotImplementedError()

    def get_package(self, package_name: str) -> importlib.metadata.Distribution | None:
        if self.site_packages_path is None:
            return None
        distributions = list(
            importlib.metadata.distributions(
                name=package_name, path=[str(self.site_packages_path)]
            )
        )
        distribution = distributions.pop(0)
        if distributions:
            raise ValueError(
                f"More than one distribution found for package {package_name} !!!"
            )
        return distribution

    def get_packages(
        self, name_filters: list[str] | None = None
    ) -> list[importlib.metadata.Distribution]:
        if self.site_packages_path is None:
            return []
        distributions = importlib.metadata.distributions(
            path=[str(self.site_packages_path)]
        )
        ret = []
        for dist in distributions:
            skip = False
            if name_filters:
                for name_filter in name_filters:
                    if name_filter not in dist.name:
                        skip = True
                        break
            if skip:
                continue
            ret.append(dist)
        return ret

    def get_packages_slow(self) -> list[tuple[str, str, str]]:
        stdout, stderr = self.get_cmd_output(
            cmd_name="uv",
            # cmd_args=["pip", "tree", "-d", "0"],
            cmd_args=["pip", "list", "--format", "json"],
        )
        try:
            data = json.loads(stdout)
        except Exception as err:
            raise ValueError(f"Error parsing pip list output: {err}")
        ret = []
        for entry in data:
            ret.append(
                (
                    entry["name"],
                    entry["version"],
                    entry.get("editable_project_location"),
                )
            )
        return ret

    def get_plugins(
        self, group_filter: str | None
    ) -> list[tuple[importlib.metadata.EntryPoint, importlib.metadata.Distribution]]:
        if self.site_packages_path is None:
            return []
        distributions = importlib.metadata.distributions(
            path=[str(self.site_packages_path)]
        )
        plugins = []
        for dist in distributions:
            for ep in dist.entry_points:
                if group_filter is None or group_filter in ep.group:
                    plugins.append([ep, dist])
        return plugins

    def get_plugins_slow(
        self, group_filter: str | None
    ) -> list[importlib.metadata.EntryPoint]:
        cmd_args = ["studio", "plugins-here", "--format", "json"]
        if group_filter:
            cmd_args.extend(["--group-filter", group_filter])
        stdout, stderr = self.get_cmd_output("tgzr", cmd_args)
        # print("???", [stdout, stderr])
        stdout = stdout.split(">>> JSON:", 1)[-1]

        data = json.loads(stdout)
        # print("-->", data)
        ret = []
        for entry in data:
            # print(entry)
            ep = importlib.metadata.EntryPoint(
                entry["name"], entry["value"], entry["group"]
            )
            ret.append(ep)
        return ret


class PackageManager:
    def __init__(self, root: Path) -> None:
        self._root = root

    @property
    def root(self) -> Path:
        return self._root

    def get_venv_path(self, venv_name: str, group: str) -> Path:
        return self.root / group / venv_name

    def get_venv(self, venv_name: str, group: str) -> Venv:
        return Venv(self.get_venv_path(venv_name, group))

    def _create_bat_shortcut(self, exe_path: Path | str, shortcut_path: Path | str):
        exe_path = Path(exe_path)
        if not exe_path.is_absolute() and not str(exe_path).startswith("./"):
            exe_path = f".\\{exe_path}"

        shortcut_path = Path(shortcut_path)
        if shortcut_path.suffix != ".bat":
            shortcut_path = f"{shortcut_path}.bat"

        content = [
            "@echo off",
            f"REM Shortcut to {exe_path}",
            "",
            f"{exe_path} %*",
        ]
        with open(shortcut_path, "w") as fp:
            fp.write("\n".join(content))

    def create_shortcut(
        self, exe_path: Path | str, shortcut_path: Path | str, relative: bool = True
    ):
        """
        Create a shortcut to exe_path at shortcut_path.

        If `relative` is True, `exe_path` will be modified to be relative
        to `shortcut_path`.

        The shortcut will be a symlink to the exe, unless the OS does not
        allow it in which case a .bat file is created.
        """
        if relative:
            exe_path = Path(exe_path).relative_to(Path(shortcut_path).parent)

        if platform.system() == "Linux":
            os.symlink(exe_path, shortcut_path, target_is_directory=False)
        else:
            self._create_bat_shortcut(exe_path, shortcut_path)

    def create_venv(
        self,
        venv_name: str,
        group: str,
        exist_ok: bool = False,
        prompt: str | None = None,
    ) -> Venv:
        venv = self.get_venv(venv_name, group)
        if venv.exists() and not exist_ok:
            raise ValueError(f"Virtual Env {venv.path} already exists!")

        venv.create(prompt, clear_existing=exist_ok)
        venv.install_uv()
        return venv



from __future__ import annotations
from typing import TYPE_CHECKING, Literal

from pathlib import Path
import importlib.metadata

from .base_config import BaseConfig, SettingsConfigDict, Field


if TYPE_CHECKING:
    from .studio import Studio


class ProjectConfig(BaseConfig):
    model_config = SettingsConfigDict(env_prefix="tgzr_project_defaults_")

    some_studio_setting: str | None = Field(
        None,
        description="This is a dummy setting. We dont have project settings yet :p",
    )
    # settings_provider: str | None = Field(
    #     None, description="Name of the settings plugin to use for this project"
    # )


class ProjectWork:
    """
    Utility class to manage the "Work" folder in a tgzr project.
    """

    def __init__(self, project: Project):
        self.project = project

    @property
    def path(self) -> Path:
        return self.project.studio.path / "WORK" / self.project.name

    @property
    def user_path(self) -> Path:
        return self.path / self.project.studio.workspace.session.context.user_name


class Project:
    @classmethod
    def config_path(cls, studio: Studio, project_name: str) -> Path:
        return studio.projects_venv_path / project_name / ".tgzr_project"

    def __init__(self, studio: Studio, name: str) -> None:
        self._studio = studio
        self._name = name

        self._config_path = self.config_path(studio, name)
        self._config = ProjectConfig.from_file(self._config_path)
        self._work = ProjectWork(self)

    @property
    def studio(self) -> Studio:
        return self._studio

    @property
    def config(self) -> ProjectConfig:
        return self._config

    @property
    def name(self) -> str:
        return self._name

    @property
    def path(self) -> Path:
        return self._studio.projects_venv_path / self._name

    @property
    def work(self) -> ProjectWork:
        # print("???", self.exists(), self._work.path.exists(), self._work.path)
        if self.exists() and not self._work.path.exists():
            self._work.path.mkdir(parents=True)
        return self._work

    @property
    def venv_path(self) -> Path:
        return self._studio.project_venv_path(self.name)

    def save_config(self) -> Path:
        """Returns the path of the saved file."""
        self._config.to_file(
            self._config_path,
            allow_overwrite=True,
            header_text=f"tgzr {self.name!r} project config",
        )
        return self._config_path

    def exists(self) -> bool:
        return self.venv_path.exists()

    def ensure_exists(self):
        if not self.exists:
            self.create()

    def create(
        self,
        required_packages: list[str] = [],
        index: str | None = None,
        find_links: str | None = None,
        allow_prerelease: bool = False,
    ) -> None:
        """
        Create (or update) this Studio and seed it with the `tgzr.shell` package.
        If `index` is None, the workspace's default index is used.
        """
        self.studio.create_project(
            self.name,
            required_packages=required_packages,
            index=index,
            find_links=find_links,
            allow_prerelease=allow_prerelease,
        )

    def get_packages(self) -> list[importlib.metadata.Distribution]:
        """
        Return a list of (package_name, version) for every package
        installed in the project.
        """
        return self.studio.get_packages(self.name)

    def get_plugins(
        self, group_filter: str | None
    ) -> list[tuple[importlib.metadata.EntryPoint, importlib.metadata.Distribution]]:
        """
        Return a list of (plugin_group, plugin_name, plugin_entry_point) for every package
        installed in the project.
        """
        return self.studio.get_plugins(self.name, group_filter)

    def run_cmd(self, cmd_name: str, cmd_args: list[str]):
        """
        Run a command in this project's venv.
        Returns True if the command was executed successfuly (according to process return int).
        """
        return self.studio.run_cmd(cmd_name, cmd_args, self.name)

    def shell(
        self,
        shell_type: Literal["xonsh", "bash", "cmd", "powershell", "auto"],
        project_name: str | None = None,
    ):
        """
        Open a shell in this project's venv.
        Returns True if the command was executed successfuly (according to process return int).
        """
        return self.studio.shell(shell_type, self.name)



from typing import Literal

import click
import rich

from tgzr.cli.utils import TGZRCliGroup

from ..workspace import Workspace
from .utils import pass_session, Session


@click.group(
    cls=TGZRCliGroup,  # to allow command short names
    help="Manage the workspace configuration",
)
def workspace():
    pass


@workspace.command("show")
@pass_session
def workspace_show(session: Session):
    """
    Show the Workspace configuration.
    """
    workspace: Workspace | None = session.workspace
    if workspace is None:
        raise click.UsageError(f"No Workspace found in {session.home}")
    rich.print(workspace.config)


@workspace.command("set")
@pass_session
@click.option(
    "--default-index",
    help=(
        "Set the url of the default package index, like "
        '"https://pypi.org/simple", "/path/to/folder" or "./path/relative/to/workspace". '
        'Use "" to unset.'
    ),
)
def workspace_set(session: Session, **kwargs):
    """
    Set some config field in the Workspace config.
    """
    workspace: Workspace | None = session.workspace
    if workspace is None:
        raise click.UsageError(f"No Workspace found in {session.home}")

    not_None_options = [n for n, v in kwargs.items() if v is not None]
    if not not_None_options:
        raise click.UsageError("Please specify at least one option.")

    # print(kwargs)

    if (default_index := kwargs["default_index"]) is not None:
        if not default_index:
            default_index = None
        workspace.config.default_index = default_index

    path = workspace.save_config()
    click.echo(f"Config saved: {path}")



import click

from tgzr.cli.utils import TGZRCliGroup

from ..session import Session, set_default_session

from .session_cli import session_group
from .workspace_cli import workspace
from .studio_cli import studio, studio_help
from .project_cli import project
from .settings_cli import settings_group
from .plugins_cli import plugins_group
from . import app_cli
from .help_cli import env_help
from .dev_cli import dev
from .utils import pass_session


def session_cwd(ctx, param, cwd):
    if cwd is None:
        cwd = "."
    session = Session(cwd)
    set_default_session(session)

    ctx.obj = session


@pass_session
def session_verbose(session: Session, ctx, param, value):
    if value:
        session.set_verbose()

    if not session.quiet:
        # NB: we do this here because it's the first callback
        # so it will be printed before anything else
        # but still is affected by -v --quiet flags
        click.echo(f"TGZR Session: home={session.home}")


@pass_session
def session_quiet(session: Session, ctx, param, value):
    if value:
        session.set_quiet()


@pass_session
def set_connection_url(session: Session, ctx, param, value):
    if value is not None:
        session.context.connection_url = value
        set_default_session(session)
        if not session.quiet:
            click.echo(f"Using connection url {session.context.connection_url}")


@pass_session
def select_user(session: Session, ctx, param, value):
    if value is not None:
        session.context.user_name = value
        set_default_session(session)
        if not session.quiet:
            click.echo(f"Using user name {session.context.user_name}")


@pass_session
def select_studio(session: Session, ctx, param, value):
    if value is not None:
        session.select_studio(value)
        set_default_session(session)
        if not session.quiet:
            click.echo(f"Using studio {session.context.studio_name}")


@pass_session
def select_project(session: Session, ctx, param, value):
    if value is not None:
        session.select_project(value)
        set_default_session(session)
        if not session.quiet:
            click.echo(f"Using project {session.context.project_name}")


def install_options(group: click.Group):
    # (see https://click.palletsprojects.com/en/stable/advanced/#id2)

    click.option(
        "-H",
        "--home",
        default=None,
        help="Folder to search for TGZR installation (default is current directory).",
        is_eager=True,
        callback=session_cwd,
        metavar="HOME",
    )(group)
    click.option(
        "-v",
        "--verbose",
        is_flag=True,
        # flag_value="",
        help="Chatty mode.",
        callback=session_verbose,
    )(group)
    click.option(
        "--quiet",
        is_flag=True,
        # flag_value="",
        help="Quiet mode (cancels --verbose).",
        callback=session_quiet,
    )(group)
    click.option(
        "-C",
        "--connection",
        default=None,
        help="Set the url to connect to (usage depends on installed connection plugins).",
        callback=set_connection_url,
        metavar="URL",
    )(group)
    click.option(
        "-U",
        "--username",
        default=None,
        help="Override user name (used in settings etc...).",
        callback=select_user,
        metavar="USER-NAME",
    )(group)
    click.option(
        "-S",
        "--studio",
        default=None,
        help="select a studio (default is defined in the workspace).",
        callback=select_studio,
        metavar="STUDIO-NAME",
    )(group)
    click.option(
        "-P",
        "--project",
        default=None,
        help="select a project (default is defined in the studio).",
        callback=select_project,
        metavar="PROJECT-NAME",
    )(group)


def install_cli(group: TGZRCliGroup):
    cmd, kwargs, setter = group.get_default_command()
    if not setter or setter.startswith("tgzr.cli"):
        # Only uninstall if the default cmd was set by tgzr.cli
        # i.e: only override if the default cmd is "install"
        # print("tgzr.shell Uninstalling default cmd from", setter)
        group.set_default_command(None)

    install_options(group)

    group.add_command(session_group)
    group.add_command(workspace)
    group.add_command(studio)
    group.add_command(project)
    group.add_command(settings_group)
    group.add_command(dev)
    group.add_command(plugins_group)

    help = group.find_group("help")
    if help is None:
        raise Exception("Could not find the cli help group :/")

    help.add_command(env_help)
    help.add_command(studio_help)

    app_cli.install_plugin(group)



from __future__ import annotations
from typing import TYPE_CHECKING, Type, TypeVar, Callable, Awaitable, Any, DefaultDict

import dataclasses
import uuid
import json
import logging

from tgzr.package_management.plugin_manager import PluginManager, Plugin

if TYPE_CHECKING:
    from .session import Session

logger = logging.getLogger(__name__)

# class BrokerImplementation:

#     def __init__(self, broker: Broker):
#         self._broker = broker

#     def publish(self, topic: str, event: Broker.Event) -> None: ...
#     def subscribe(
#         self, topic_pattern: str, callback: EventCallbackType
#     ) -> Broker.Subscription: ...
#     def unsubscribe(self, subscription: Broker.Subscription) -> bool: ...


# class TestBroker(BrokerImplementation):
#     def __init__(self, broker: Broker):
#         super().__init__(broker=broker)
#         self._subscriptions: DefaultDict[str, list[Broker.Subscription]] = defaultdict(
#             list
#         )

#     def topic_match(self, topic: str, topic_pattern: str):
#         # TODO: implement it like nats or zenoh
#         return fnmatch.fnmatch(topic, topic_pattern)

#     def publish(self, topic: str, event: Broker.Event) -> None:
#         print("Publish", repr(topic), event)
#         for topic_pattern, subscriptions in self._subscriptions.items():
#             # print("?? topic match?", topic_pattern)
#             if self.topic_match(topic, topic_pattern):
#                 for subscription in subscriptions:
#                     print(
#                         f"  --push--> event to {subscription.topic_pattern!r}: {subscription.callback}"
#                     )
#                     subscription.callback(event)

#     def subscribe(
#         self, topic_pattern: str, callback: EventCallbackType
#     ) -> Broker.Subscription:
#         subscription = self._broker.Subscription(
#             topic_pattern=topic_pattern,
#             callback=callback,
#         )
#         self._subscriptions[topic_pattern].append(subscription)
#         return subscription

#     def unsubscribe(
#         self, subscription: Broker.Subscription, raise_if_unknown: bool = True
#     ) -> bool:
#         topic_pattern = subscription.topic_pattern
#         subscriptions = self._subscriptions[topic_pattern]
#         try:
#             subscriptions.remove(subscription)
#         except ValueError:
#             if not raise_if_unknown:
#                 return False
#             raise ValueError(f"This subscription is not registerd here: {subscription}")
#         return True


# class Broker:

#     class Event:
#         def __init__(self, **data: Any):
#             self._data = data

#         def __str__(self) -> str:
#             kwargs = ", ".join([f"{k}={v!r}" for k, v in self._data.items()])
#             return f"{self.__class__.__name__}({kwargs})"

#         def unpack(self, *args):
#             data = self._data.copy()
#             ret = []
#             for name in args:
#                 try:
#                     ret.append(data.pop(name))
#                 except KeyError:
#                     raise KeyError(
#                         f"Cannot unpack {args} from {self._data}, missing {name!r}."
#                     )
#             ret.append(data)
#             return ret

#     @dataclasses.dataclass
#     class Subscription:
#         topic_pattern: str
#         callback: Callable[[Broker.Event], None]
#         uuid: uuid.UUID = dataclasses.field(default_factory=uuid.uuid4)

#     def __init__(self, session: Session):
#         self._session = session
#         self._broker_implementation = self._create_broker_implementation()

#     def _create_broker_implementation(self) -> BrokerImplementation:
#         # TODO: select the Broker class from self.session.config
#         broker_implementation = TestBroker(self)
#         return broker_implementation

#     def publish(self, topic: str, **data: Any) -> None:
#         self._broker_implementation.publish(topic, self.Event(**data))

#     def cmd(self, cmd_name: str, **kwargs) -> None:
#         topic = "$CMD." + cmd_name
#         event = self.Event(**kwargs)
#         self._broker_implementation.publish(topic, event=event)

#     def subscribe(
#         self, topic_pattern: str, callback: EventCallbackType
#     ) -> Broker.Subscription:
#         return self._broker_implementation.subscribe(
#             topic_pattern=topic_pattern, callback=callback
#         )

#     def unsubscribe(self, subscription: Broker.Subscription) -> bool:
#         return self._broker_implementation.unsubscribe(subscription=subscription)


class AsyncBrokerImplementation(Plugin):

    @classmethod
    def plugin_type_name(cls) -> str:
        return "AsyncBrokerImplementation"

    async def connect(self, config: dict[str, Any]): ...
    def connected(self) -> bool: ...
    async def disconnect(self): ...
    async def publish(self, subject: str, event: AsyncBroker.Event) -> None: ...
    async def query(
        self, subject: str, event: AsyncBroker.Event, timeout: float = 0.5
    ) -> AsyncBroker.Event: ...
    async def subscribe(
        self,
        subject_pattern: str,
        callback: Callable[[AsyncBroker.Event], Awaitable[None]],
    ) -> AsyncBroker.Subscription: ...
    async def unsubscribe(self, subscription: AsyncBroker.Subscription) -> bool: ...


class BrokerPluginsManager(PluginManager[AsyncBrokerImplementation]):
    EP_GROUP = "tgzr.shell.broker_plugin"


EventType = TypeVar("EventType", bound="AsyncBroker.Event")


class AsyncBroker:
    class ExtendedJSONEncoder(json.JSONEncoder):
        def default(self, obj):
            if isinstance(obj, uuid.UUID):
                return str(obj)
            return json.JSONEncoder.default(self, obj)

    class Event:
        @classmethod
        def from_bytes(cls: Type[EventType], data: bytes) -> EventType:
            py_data = json.loads(data.decode())
            event = cls(py_data)
            return event

        def __init__(self, data: Any):
            self._data = data

        def __str__(self) -> str:
            # kwargs = ", ".join([f"{k}={v!r}" for k, v in self._data.items()])
            # return f"{self.__class__.__name__}({kwargs})"
            return f"{self.__class__.__name__}({self._data!r})"

        @property
        def data(self) -> Any:
            return self._data

        def unpack(self, *args):
            data = self._data.copy()
            ret = []
            for name in args:
                try:
                    ret.append(data.pop(name))
                except KeyError:
                    raise KeyError(
                        f"Cannot unpack {args} from {self._data}, missing {name!r}."
                    )
            ret.append(data)
            return ret

        def to_bytes(self) -> bytes:
            return json.dumps(self._data, cls=AsyncBroker.ExtendedJSONEncoder).encode()

    @dataclasses.dataclass
    class Subscription:
        subject_pattern: str
        callback: Callable[[AsyncBroker.Event], None]
        uuid: uuid.UUID = dataclasses.field(default_factory=uuid.uuid4)
        private_data: Any | None = None

    def __init__(self, session: Session):
        self._plugin_manager = BrokerPluginsManager()

        self._session = session
        self._broker_implementation = self._create_broker_implementation()

    def _create_broker_implementation(self) -> AsyncBrokerImplementation:
        broker_plugin_name = "tgzr.shell_brokers.jsbroker.JSBroker"
        async_broker_implementation = self._plugin_manager.get_plugin(
            broker_plugin_name
        )
        return async_broker_implementation

    def connected(self) -> bool:
        return self._broker_implementation.connected()

    async def connect(self, config: dict[str, Any]):
        await self._broker_implementation.connect(config)

    async def disconnect(self):
        await self._broker_implementation.disconnect()

    async def publish(self, subject: str, **data: Any) -> None:
        await self._broker_implementation.publish(subject, self.Event(data))

    # async def request(
    #     self, subject: str, event: AsyncBroker.Event, timeout: float = 0.5
    # ) -> Any:
    #     await self._broker_implementation.request(subject, event, timeout)

    async def cmd(self, service_name: str, cmd_name: str, **kwargs) -> None:
        subject = service_name + ".$CMD." + cmd_name
        event = self.Event(kwargs)
        await self._broker_implementation.publish(subject, event=event)

    async def query(
        self, service_name: str, query_name: str, timeout: float = 0.5, **kwargs
    ) -> None:
        subject = service_name + ".$QUERY." + query_name
        event = self.Event(kwargs)
        logger.debug(f"[SENDING QUERY] {subject=}, {event=}, {timeout=}")
        event = await self._broker_implementation.query(
            subject, event=event, timeout=timeout
        )
        logger.debug(f"[QUERY SENT] {query_name}, {kwargs} @ {subject} -> {event}")
        return event.data

    async def subscribe(
        self,
        subject_pattern: str,
        callback: Callable[[AsyncBroker.Event], Awaitable[None]],
    ) -> AsyncBroker.Subscription:
        return await self._broker_implementation.subscribe(
            subject_pattern=subject_pattern, callback=callback
        )

    async def unsubscribe(self, subscription: AsyncBroker.Subscription) -> bool:
        return await self._broker_implementation.unsubscribe(subscription=subscription)


# def test():
#     from .session import get_default_session
#     import os

#     os.environ["tgzr_home"] = "/home/dee/DEV/_OPEN-TGZR_/TGZR"
#     session = get_default_session(ensure_set=True)
#     if session is None:
#         raise Exception("No session!")
#     print(session.broker._broker_implementation)


# if __name__ == "__main__":
#     test()



from __future__ import annotations
from typing import TYPE_CHECKING

from pathlib import Path

from .base_config import BaseConfig, SettingsConfigDict, Field
from .studio import Studio

if TYPE_CHECKING:
    from .session import Session


class WorkspaceConfig(BaseConfig):
    model_config = SettingsConfigDict(env_prefix="tgzr_workspace_")

    default_studio_name: str | None = Field(
        None, description="Name of the default studio"
    )
    default_index: str | None = Field(
        None,
        description="Default package index for all studios. Use None for https://pypi.org/simple",
    )


class Workspace:
    def __init__(self, session: Session):
        if not session.home.is_dir():
            raise FileNotFoundError(
                f'The path "{self._path}" is not a directory, cannot use it as Workspace parent.'
            )
        self.session = session
        self._path = (session.home / "Workspace").resolve()
        if not self._path.exists():
            self._path.mkdir()
        self._config_path = self._path / ".tgzr_workspace"
        self._config = WorkspaceConfig.from_file(self._config_path)

    @property
    def path(self) -> Path:
        return self._path

    @property
    def config(self) -> WorkspaceConfig:
        return self._config

    def studio_names(self) -> list[str]:
        """
        Return the names of the installed studios.
        """
        if not self._path.exists():
            return []
        names = []
        for path in self._path.iterdir():
            config_path = Studio.config_path(self, path.name)
            if config_path.exists():
                names.append(path.name)
        return sorted(names)

    def get_studios(self) -> list[Studio]:
        return [
            self.get_studio(name, ensure_exists=False) for name in self.studio_names()
        ]

    def set_default_studio(
        self,
        studio_name: str | None,
        ensure_exists: bool = True,
        save_config: bool = True,
    ) -> None:
        if studio_name is None:
            self._config.default_studio_name = None
        else:
            if ensure_exists:
                # will raise if studio does not exists on disk:
                self.get_studio(studio_name, ensure_exists=True)
            self._config.default_studio_name = studio_name
        if save_config:
            self.save_config()

    def save_config(self) -> Path:
        """Returns the path of the saved file."""
        self._config.to_file(
            self._config_path,
            allow_overwrite=True,
            header_text="tgzr workspace config",
        )
        return self._config_path

    def default_studio_name(self) -> str | None:
        """The name of the default studio, or None"""
        return self.config.default_studio_name

    def get_default_studio(self, ensure_exists: bool = True) -> Studio | None:
        name = self._config.default_studio_name
        if name is None:
            return None
        return self.get_studio(name, ensure_exists)

    def get_studio(self, name: str, ensure_exists: bool = True) -> Studio:
        """
        Return the studio with name "name".
        If ensensure_exists and the studio does not exist: raise FileNotFoundError.
        """
        studio = Studio(self, name)
        if ensure_exists and not studio.exists():
            raise FileNotFoundError
        return studio

    def resolve_index(
        self, index: str | None, relative_to: Path | None = None
    ) -> str | None:
        """
        Returns the path of an index, making it absolute in case
        of relative local path.
        If `relative_to` is not provided, `self.path` is used.

        If index is None, the default index will be returned.
        If no default index is set, None is returned.
        """
        index = index or self.config.default_index
        if index is None:
            return None
        if not "://" in index:
            local_index = Path(index)
            if not local_index.is_absolute():
                relative_to = relative_to or self.path
                index = str((relative_to / local_index).resolve())
        return index



import click

from tgzr.cli.utils import TGZRCliGroup

from tgzr.cli.add_plugins import BrokenCommand

from ..app_sdk import run_app_process
from ..app import get_apps, get_broken_apps
from ..app_sdk._base_app import _BaseShellApp
from ..app_sdk.nice_app import ShellNiceApp


@click.group("app", cls=TGZRCliGroup, help="Run installed applications.")
def app_group():
    pass


def create_nice_run_command(app: ShellNiceApp) -> click.Command:
    @click.command(app.app_name)
    @click.option(
        "-r",
        "--reload",
        is_flag=True,
        help="Reload on code change (will run in browser).",
    )
    @click.option(
        "-d",
        "--detach",
        is_flag=True,
        help="Launch the App in another process to avoid blocking this one.",
    )
    def run_app(reload: bool, detach: bool):
        try:
            run_app_process(app, detach=detach, notify=click.echo, reload=reload)
        except ValueError as err:
            raise click.UsageError(str(err))

    return run_app


def create_run_command(app: _BaseShellApp) -> click.Command:
    @click.command(app.app_name)
    @click.option(
        "-d",
        "--detach",
        is_flag=True,
        help="Launch the App in another process to avoid blocking this one.",
    )
    def run_app(detach: bool):
        try:
            run_app_process(app, detach=detach, notify=click.echo)
        except ValueError as err:
            raise click.UsageError(str(err))

    return run_app


def install_plugin(group: TGZRCliGroup):
    broken_apps = get_broken_apps()
    for entry_point, exception in broken_apps:
        # print("Broken:", entry_point)
        app_group.add_command(BrokenCommand(entry_point, exception))

    apps = get_apps()
    for app in apps:
        # print("Adding app to cli:", app.app_name, app)
        if isinstance(app, ShellNiceApp):
            cmd = create_nice_run_command(app)
        else:
            cmd = create_run_command(app)
        app_group.add_command(cmd)
        app.cli_run_cmd_installed(cmd, group)

    group.add_command(app_group)

    return group



from __future__ import annotations
from typing import TYPE_CHECKING, Callable, Any

import pydantic

from . import run_app_process

from ..session import Session


class ShellAppContext(pydantic.BaseModel):
    # FIXME: this should be a dataclass, no need for pydantic here.
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    session: Session
    host_name: str
    context_name: str


class DefaultShellAppInfo(pydantic.BaseModel):
    title: str | None = pydantic.Field(
        default=None,
        description="The title to show. If not provided, a prettyfied version of the app name will be used.",
    )
    icon: str | None = pydantic.Field(
        default=None,
        description="A font-awesome or google font icon name",
        examples=["fa-brands fa-artstation", "stat_shine"],
    )
    # TODO: find a way to specify color in #FFFFFF so qt app can use it.
    color: str | None = pydantic.Field(
        default=None,
        description="The color to use when showing the app in UIs (see https://quasar.dev/style/color-palette#color-list)",
    )

    # TODO: find out if there's interest in have those in defaults two:
    # needs_dialog: bool = False
    # enabled: bool = True
    # hidden: bool = False

    # installed_versions: set[str] = set()


class ShellAppInfo(pydantic.BaseModel):
    # model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    app: object  # pointer to the app TODO: find a more elegant way to fix the import loop.
    title: str
    icon: str | None = pydantic.Field(
        default=None,
        description="A font-awesome or google font icon name",
        examples=["fa-brands fa-artstation", "stat_shine"],
    )
    color: str | None = None

    needs_dialog: bool = False
    enabled: bool = True
    hidden: bool = False

    installed_versions: set[str] = set()

    @property
    def app_id(self) -> str:
        return self.app.app_id  # type: ignore Trust Me Sis©

    @property
    def groups(self) -> set[str]:
        return self.app.app_groups.copy()  # type: ignore Trust Me Sis©

    def run_app(
        self,
        detach: bool = True,
        notify: Callable[[str], None] | None = None,
        **options: dict[str, Any],
    ):
        run_app_process(
            app=self.app,  # type: ignore Trust Me Sis©
            detach=detach,
            notify=notify,
            **options,
        )



from __future__ import annotations
from typing import TYPE_CHECKING, Callable, Any

import sys
import os
import subprocess

if TYPE_CHECKING:
    from ._base_app import _BaseShellApp
    from .nice_app import ShellNiceApp


def run_app_process(
    app: _BaseShellApp,
    detach: bool = True,
    notify: Callable[[str], None] | None = None,
    **options: Any,
):
    if notify is None:
        notify = print

    from .nice_app import ShellNiceApp

    if isinstance(app, ShellNiceApp):
        launcher_module = app.run_native_module.__name__
        if options.get("reload"):
            if detach:
                raise ValueError(
                    "You can't run detached with reload (you wouldn't be able to close the process). "
                    "Please choose one or the other."
                )
            launcher_module = app.run_dev_module.__name__
    else:
        launcher_module = app.run_native_module.__name__

    cmd = [sys.executable, "-m", launcher_module]
    if detach:
        # TODO: implement a subprocess manager to manage popen instances.
        popen = subprocess.Popen(cmd)
        notify(f"New Subprocess: {popen}")
    else:
        cmd = " ".join(cmd)
        notify(f"Executing System Command: {cmd}")
        os.system(cmd)



from typing import Literal


import click
import rich
import rich.table

from tgzr.cli.utils import TGZRCliGroup

from ..session import Session
from ..workspace import Workspace
from ..studio import Studio

from .utils import pass_session


@click.command("studio")
def studio_help():
    rich.print("Sorry, this help is not written yet :p")


@click.group(cls=TGZRCliGroup, help="Manage tgzr studios")
def studio():
    pass


@studio.command()
@click.argument("name")
@click.option("--allow-existing", is_flag=True, default=False)
@click.option(
    "--default-index",
    help="The URL of the default package index (by default: <https://pypi.org/simple>).",
)
@click.option(
    "-f",
    "--find-links",
    help="path a folder containing packages to install. Usefull in no-internet situations.",
)
@click.option(
    "--allow-prerelease",
    is_flag=True,
    help="Allow installing tgzr using pre-release packages. Default is False.",
)
@click.option(
    "-N",
    "--not-default",
    is_flag=True,
    default=False,
    help="Do not make this created studio the default one.",
)
@pass_session
def create(
    session: Session,
    name,
    allow_existing: bool = False,
    default_index: str | None = None,
    find_links: str | None = None,
    allow_prerelease: bool = False,
    not_default: bool = True,
):
    """
    Create a new Studio with name "NAME".
    """
    workspace: Workspace | None = session.workspace
    if workspace is None:
        raise click.UsageError(f"No workspace found at {session.home}.")
    studio = workspace.get_studio(name, ensure_exists=False)
    if studio.exists():
        if not allow_existing:
            raise click.UsageError(
                f'A Studio "{name}" already exists in workspace "{workspace.path}". Use --allow-existing to update it.'
            )

    studio.create(
        index=default_index,
        find_links=find_links,
        allow_prerelease=allow_prerelease,
    )
    click.echo(f"Studio {name!r} created.")

    # Ensure the studio config is written so the session's home can
    # be discovered:
    session.save_config()

    if not_default:
        return

    workspace.set_default_studio(name)
    click.echo(f'Studio "{name}" set as default studio.')


@studio.command()
@pass_session
def ls(session: Session):
    """
    List the installed Studio.
    """
    workspace: Workspace | None = session.workspace
    if workspace is None:
        raise click.UsageError(f"No workspace found at {session.home}.")
    default_studio_name = workspace.default_studio_name()
    for studio in workspace.get_studios():
        default_indic = ""
        if studio.name == default_studio_name:
            default_indic = "*"
        click.echo(f"  {studio.name}{default_indic} ({studio.path})")


@studio.command()
@pass_session
def show(session: Session):
    """Show the studio config."""
    try:
        studio: Studio | None = session.get_selected_studio()
    except FileNotFoundError:
        raise click.UsageError(
            f'Studio "{session.selected_studio_name}" does not exists.'
        )
    if studio is None:
        raise click.UsageError(
            "No Studio selected, use `tgzr --studio <name> ...` to specify one, or `tgzr studio select` to set the default one."
        )

    click.echo(f'Configuration for Studio "{studio.name}" ({studio.path}):')
    rich.print(studio.config)


@studio.command()
@pass_session
def select(session: Session):
    """
    Set the default studio (without -S or --studio, resets default to None).
    """
    workspace: Workspace | None = session.workspace
    if workspace is None:
        raise click.UsageError(f"No workspace found at {session.home}.")

    default_studio_name = workspace.default_studio_name()
    studio: Studio | None = session.get_selected_studio()
    if studio is None:
        workspace.set_default_studio(None)
    else:
        if studio.name == default_studio_name:
            click.echo(f"The default Studio is already {default_studio_name}.")
        else:
            workspace.set_default_studio(studio.name)


@studio.command()
@pass_session
def switch(session: Session):
    """
    Change the default studio.
    """
    workspace: Workspace | None = session.workspace
    if workspace is None:
        raise click.UsageError(f"No workspace found at {session.home}.")

    studio_names = workspace.studio_names()
    if not studio_names:
        click.echo('No Studio installed yet. Use "tgzr studio create" to add one.')
        return

    click.echo("0 <None>")
    for i, name in enumerate(studio_names, start=1):
        click.echo(f"{i} {name}")
    value = click.prompt("Enter Studio")
    if value.strip() == "0":
        workspace.set_default_studio(None)
        return

    name = None
    try:
        index = int(value)
    except ValueError:
        if value in studio_names:
            name = value
    else:
        try:
            name = studio_names[index - 1]
        except:
            pass

    if name is None:
        click.echo("Nope ¯\\_(ツ)_/¯")
    else:
        workspace.set_default_studio(name)


@studio.command()
@pass_session
@click.option(
    "--default-index",
    help=(
        "Set the url of the default package index, like "
        '"https://pypi.org/simple", "/path/to/folder" or "./path/relative/to/workspace". '
        'Use "" to unset.'
    ),
)
def set(session: Session, **kwargs):
    """
    Set some config field in the Studio config.
    """
    studio: Studio | None = session.get_selected_studio()
    if studio is None:
        raise click.UsageError("Please select a studio or set a default one.")

    not_None_options = [n for n, v in kwargs.items() if v is not None]
    if not not_None_options:
        raise click.UsageError("Please specify at least one option.")

    # print(kwargs)

    if (default_index := kwargs["default_index"]) is not None:
        if not default_index:
            default_index = None
        studio.config.default_index = default_index

    path = studio.save_config()
    click.echo(f"Config saved: {path}")


@studio.command(
    context_settings=dict(
        ignore_unknown_options=True,
    )
)
@click.argument(
    "cmd_name",
)
@click.argument("cmd_args", nargs=-1, type=click.UNPROCESSED)
@pass_session
def run(session: Session, cmd_name: str, cmd_args: list[str]):
    """
    Run a studio cmd.
    """
    studio: Studio | None = session.get_selected_studio()
    if studio is None:
        raise click.UsageError("Please select a studio or set a default one.")

    studio.run_cmd(cmd_name, cmd_args, project_name=None)


@studio.command
@click.option(
    "-s",
    "--shell",
    default="auto",
    help='The shell to use, one of ["xonsh", "bash", "cmd", "powershell", "auto"] (defaults to "auto").',
)
@pass_session
def shell(
    session: Session,
    shell: Literal["xonsh", "cmd", "powershell", "auto"] = "auto",
):
    """
    Open a shell configured for the studio.
    """
    studio: Studio | None = session.get_selected_studio()
    if studio is None:
        raise click.UsageError("Please select a studio or set a default one.")
    studio.shell(shell_type=shell, project_name=None)


@studio.command()
@click.option(
    "-f",
    "--group-filter",
    help="Show only this type of plugins (defaults is to show all).",
)
@pass_session
def plugins(session: Session, group_filter: str | None):
    """Show the plugins installed in the studio venv."""
    try:
        studio: Studio | None = session.get_selected_studio()
    except FileNotFoundError:
        raise click.UsageError(
            f'Studio "{session.selected_studio_name}" does not exists.'
        )
    if studio is None:
        raise click.UsageError(
            "No Studio selected, use `tgzr --studio <name> ...` to specify one, or `tgzr studio select` to set the default one."
        )

    plugins = studio.get_plugins(None, group_filter)
    table = rich.table.Table(
        "Group",
        "Name",
        "Value",
        title=f'Plugins in Studio "{studio.name}" ({studio.path})',
        min_width=80,
    )
    for plugin, distribution in plugins:
        package = f"{distribution.name} {distribution.version}"
        editable = distribution.origin and distribution.origin.dir_info.editable
        if editable:
            package = f"[orange1]{package}[/orange1]"
        table.add_row(
            plugin.group,
            plugin.name,
            plugin.value,
            package,
        )
    rich.print(table)



from __future__ import annotations
from typing import TYPE_CHECKING

from .settings_provider import BaseSettingsProvider
from .settings_provider import (
    SettingsModelType,
)  # for transitiv imports

if TYPE_CHECKING:
    from ..session import Session


class SettingsProviderPlugin:

    def __init__(self, session: Session) -> None:
        pass

    def handles(self, url: str) -> bool:
        return True

    def create_provider(self, url: str) -> BaseSettingsProvider:
        raise NotImplementedError()



from __future__ import annotations
from typing import TYPE_CHECKING, get_args, Type, TypeVar, Callable, Any, Iterable

import importlib.metadata
import inspect
import logging

T = TypeVar("T", bound="Plugin")

logger = logging.getLogger(__name__)


class Plugin:
    """
    Subclass to create your plugin type.

    NOTE: subclass should not change __init__ signature and always
    pass *args and **kwargs to super().__init__

    """

    @classmethod
    def plugin_type_name(cls) -> str:
        raise NotImplementedError()

    @classmethod
    def plugin_name(cls) -> str:
        return f"{cls.__module__}.{cls.__name__}"

    @classmethod
    def plugin_id(cls) -> str:
        return f"{cls.plugin_type_name()}@{cls.plugin_name()}"

    def __init__(self, ep: importlib.metadata.EntryPoint):
        super().__init__()
        self._entry_point = ep

    def plugin_info(self) -> dict[str, Any]:
        return dict(
            plugin_type_name=self.plugin_type_name(),
            plugin_name=self.plugin_name(),
            plugin_id=self.plugin_id(),
            entry_point=self._entry_point,
        )


class PluginManager[PluginType: Plugin]:
    EP_GROUP = "your_plugin_entry_point_group"

    @classmethod
    def managed_plugin_type(cls) -> Type[PluginType]:
        return get_args(cls.__orig_bases__[0])[0]  # type: ignore __orig_bases__ trust me bro.

    def __init__(self):
        self._broken: list[tuple[importlib.metadata.EntryPoint, Exception]] = []
        self._loaded: list[PluginType] = []
        self._needs_loading: bool = True

    def _instantiate_plugin_type(
        self, PluginType: Type[PluginType], entry_point: importlib.metadata.EntryPoint
    ) -> PluginType:
        """
        Subclasses will want to override this if the managed plugin type needs
        init args / kwargs.
        """
        return PluginType(entry_point)

    def _resolve_plugins(
        self,
        loaded: (
            PluginType
            | Type[PluginType]
            | Callable[[], PluginType | list[PluginType]]
            | Iterable[PluginType]
        ),
        entry_point: importlib.metadata.EntryPoint,
    ) -> list[PluginType]:
        # print("Resolving shell app plugins:", loaded)
        ManagedPluginType = self.__class__.managed_plugin_type()

        if isinstance(loaded, ManagedPluginType):
            # print("  is app")
            return [loaded]

        elif inspect.isclass(loaded) and issubclass(loaded, ManagedPluginType):
            # print("  is a plugin type")
            return [self._instantiate_plugin_type(loaded, entry_point)]

        elif callable(loaded):
            # print("  is callable (but not a plugin type)")
            try:
                plugin_or_list_of_plugins = loaded()  # type: ignore
            except Exception as err:
                ValueError(
                    f"Error while executing callable entry point value (ep={entry_point}): {err}"
                )
            return self._resolve_plugins(
                loaded=plugin_or_list_of_plugins, entry_point=entry_point
            )

        elif isinstance(loaded, (tuple, list, set)):
            # print("  is iterable")
            return [plugin for plugin in loaded]

        # print("  is unsupported")
        raise ValueError(
            f'Invalid value for "{self.EP_GROUP}" entry point. '
            f"Must be a {ManagedPluginType}, a list/tuple/set of {ManagedPluginType}, or a callable returning one of these"
        )

    def _load_plugins(self):
        all_entry_points = importlib.metadata.entry_points(group=self.EP_GROUP)

        self._broken.clear()
        self._loaded.clear()

        for ep in all_entry_points:
            logger.info(f"Loading {self.EP_GROUP} plugin:", ep.name)
            try:
                loaded = ep.load()
            except Exception as err:
                self._broken.append((ep, err))
            else:
                try:
                    plugins = self._resolve_plugins(loaded, ep)
                except Exception as err:
                    self._broken.append((ep, err))
                else:
                    for plugin in plugins:
                        self._loaded.append(plugin)

        self._needs_loading = False

    def get_plugins(self, force_reload: bool = False) -> list[PluginType]:
        if force_reload or self._needs_loading:
            self._load_plugins()
        return self._loaded

    def get_plugin(self, plugin_name: str) -> PluginType:
        for plugin in self.get_plugins():
            if plugin.plugin_name() == plugin_name:
                return plugin
        raise ValueError(
            f"Not {self.managed_plugin_type().plugin_type_name()} plugin found with name {plugin_name!r}. "
            f"(got plugins:{[p.plugin_name() for p in self.get_plugins()]} and errors: {self._broken})"
        )

    def find_plugins(
        self, PluginType: Type[T], raise_not_found: bool = True
    ) -> list[T]:
        found = []
        for plugin in self.get_plugins():
            if isinstance(plugin, PluginType):
                found.append(plugin)
        if not found:
            raise ValueError(
                f"No plugin with type {PluginType} found!."
                f"(got plugins:{[p.plugin_name() for p in self.get_plugins()]} and errors: {self._broken})"
            )
        return found


def test():
    class MyPlugin(Plugin):
        pass

    class MyPluginManager(PluginManager[MyPlugin]):
        EP_GROUP = "my_plugin_entry_point_group"

    pm = MyPluginManager()
    print(pm.managed_plugin_type())


if __name__ == "__main__":
    test()



from __future__ import annotations
from typing import TYPE_CHECKING, Type, Any

from tgzr.package_management.plugin_manager import PluginManager, Plugin


if TYPE_CHECKING:
    from .session import Session
    from .broker import AsyncBroker  # used for type checking transitive imports


class ServiceClientPlugin(Plugin):
    @classmethod
    def plugin_type_name(cls) -> str:
        return "ServiceClientPlugin"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._session: Session | None = None

    @property
    def session(self) -> Session:
        if self._session is None:
            raise ValueError("Not connected! Did you call sesssion.connect() ?")
        return self._session

    async def connect(self, session: Session) -> None:
        self._session = session

    async def disconnect(self) -> None: ...


class ServiceClientPluginManager(PluginManager[ServiceClientPlugin]):
    EP_GROUP = "tgzr.shell.service_client_plugin"



from __future__ import annotations
from typing import Type, TypeVar

from pathlib import Path

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

SettingsConfigDict = SettingsConfigDict  # used as transitive import
Field = Field  # used as transitive import

ConfigType = TypeVar("ConfigType", bound="BaseConfig")


class BaseConfig(BaseSettings):

    @classmethod
    def find_config_file(cls, directory: Path, name: str) -> Path | None:
        for parent in [directory] + list(directory.parents):
            config_path = parent / name
            if config_path.exists():
                return config_path
            for sub in sorted(parent.iterdir()):
                if sub.is_dir():
                    config_path = sub / name
                    try:
                        exists = config_path.exists()
                    except PermissionError:
                        exists = False
                    if exists:
                        return config_path
        return None

    @classmethod
    def used_env_vars(cls) -> list[tuple[str, str | None]]:
        env_vars = []
        prefix = cls.model_config.get("env_prefix", "")
        for name, field in cls.model_fields.items():
            env_name = f"{prefix}{name}"
            env_vars.append((env_name, field.description))
        return env_vars

    @classmethod
    def from_file(
        cls: Type[ConfigType], path: str | Path, ensure_exists: bool = False
    ) -> ConfigType:
        path = Path(path)
        if ensure_exists and not path.exists():
            raise FileExistsError(path)
        config = cls(_env_file=path.resolve())  # type: ignore
        return config

    def to_file(
        self, path: str | Path, allow_overwrite: bool = False, header_text: str = ""
    ):
        header_text = header_text or "tgzr config file"
        lines = [
            f"# {header_text}",
            "",
        ]
        prefix = self.model_config.get("env_prefix", "")
        for name, value in self.model_dump(
            exclude_unset=True,
            exclude_defaults=True,
            exclude_none=True,
            exclude_computed_fields=True,
        ).items():
            lines.append(f"{prefix}{name}={value!r}")

        content = "\n".join(lines)

        path = Path(path)
        if path.exists() and not allow_overwrite:
            raise FileExistsError(path.resolve())
        else:
            path.parent.mkdir(parents=True, exist_ok=True)

        with open(path, "w") as fp:
            fp.write(content)



from __future__ import annotations
from typing import TYPE_CHECKING, Any, TypeVar, Callable, Awaitable

import pydantic


from .services import ServiceClientPlugin

if TYPE_CHECKING:
    from .services import AsyncBroker

SettingsModelType = TypeVar("SettingsModelType", bound=pydantic.BaseModel)


class SettingsClientPlugin(ServiceClientPlugin):
    @classmethod
    def plugin_type_name(cls) -> str:
        return "SettingsClientPlugin"

    #
    # ---
    #
    async def watch_changes(
        self, callback: Callable[[AsyncBroker.Event], Awaitable[None]]
    ): ...

    #
    # ---
    #

    async def get_context_names(self) -> tuple[str, ...]: ...
    async def set_context_info(self, context_name: str, **info: Any) -> None: ...
    async def get_context_info(self, context_name: str) -> dict[str, Any]: ...
    def expand_context_name(self, context_name: str) -> list[str]: ...

    #
    # ---
    #
    async def get_context_flat(
        self,
        context: list[str],
        path: str | None = None,
        with_history: bool = False,
    ) -> dict[str, Any]: ...

    async def get_context_dict(
        self, context: list[str], path: str | None = None, with_history: bool = False
    ) -> dict[str, Any]: ...

    async def get_context(
        self,
        context: list[str],
        model_type: type[SettingsModelType],
        path: str | None = None,
    ) -> SettingsModelType: ...

    #
    # ---
    #
    async def update_context_flat(
        self, context_name: str, flat_dict: dict[str, Any], path: str | None = None
    ) -> None: ...

    async def update_context_dict(
        self,
        context_name: str,
        deep_dict: dict[str, Any | dict[str, Any]],
        path: str | None = None,
    ) -> None: ...

    async def update_context(
        self,
        context_name: str,
        model: SettingsModelType,
        path: str | None = None,
        exclude_defaults: bool = True,
    ): ...

    #
    # ---
    #

    async def set(self, context_name: str, name: str, value: Any) -> None: ...
    async def toggle(self, context_name: str, name: str) -> None: ...
    async def add(self, context_name: str, name: str, value: Any) -> None: ...
    async def sub(self, context_name: str, name: str, value: Any) -> None: ...
    async def set_item(
        self, context_name: str, name: str, index: int, item_value: Any
    ) -> None: ...
    async def del_item(self, context_name: str, name: str, index: int) -> None: ...
    async def remove(self, context_name: str, name: str, item: str) -> None: ...
    async def append(self, context_name: str, name: str, value: Any) -> None: ...
    async def env_override(
        self, context_name: str, name: str, envvar_name: str
    ) -> None: ...
    async def pop(self, context_name: str, name: str, index: int | slice) -> None: ...
    async def remove_slice(
        self,
        context_name: str,
        name: str,
        start: int,
        stop: int,
        step: int | None = None,
    ) -> None: ...
    async def call(
        self,
        context_name: str,
        name: str,
        method_name: str,
        args: list[Any],
        kwargs: dict[str, Any],
    ) -> None: ...



from __future__ import annotations
from typing import TYPE_CHECKING
from types import ModuleType

from .exe_app import ShellExeApp as _ShellExeApp
from .exe_app import DefaultShellAppInfo

if TYPE_CHECKING:
    from ..session import Session


class ShellHostApp(_ShellExeApp):
    """
    This special type of ShellExeApp
    represents an "Integrated DCC".
    i.e: a DCC which can host a TGZR session
    and use `tgzr.host.plugin` plugins.

    Subclass will typically override these methods:
    - installable_versions(session)
    - install_version(session, version_name)
    - installed_versions()
    - env(session, version)
    - exe_path(session, version)
    - cmd(session, version)

    See docstrings in ShellHostApp and its base class for
    details on each of them.
    """

    def __init__(
        self,
        host_name: str,
        run_module: ModuleType | None,
        app_id: str | None = None,
        app_groups: set[str] = set(),
        default_app_info: DefaultShellAppInfo | None = None,
    ):
        app_groups.add("_HOST_")
        app_groups.add("host")
        super().__init__(
            exe_path=None,
            app_name=host_name,
            run_module=run_module,
            app_id=app_id,
            app_groups=app_groups,
            default_app_info=default_app_info,
        )
        self.app_groups.remove("_EXE_")

    def exe_path(self, session: Session, version: str) -> str:
        """
        Subclass need to implement this to
        locate the appropriate exe path and return it.
        """
        raise NotImplementedError()



from pathlib import Path
import ast
import asyncio

import click
import rich
import rich.table
import json

from tgzr.cli.utils import TGZRCliGroup

from .utils import pass_session, Session


@click.group(cls=TGZRCliGroup, help="Manage tgzr settings.")
def settings_group():
    pass


@settings_group.command()
@pass_session
def show(session: Session):
    rich.print(session.settings.plugin_info())


@settings_group.command()
@click.argument("key")
@click.argument("context", nargs=-1, required=True)
@click.option(
    "-f",
    "--format",
    default="text",
    help='The output format ["text", "json"] (defaults to "text").',
)
@click.option(
    "--default",
    help="The value to show if the key is not set in the given context.",
)
@pass_session
def get(session: Session, context, key, format, default):
    if default == "None":
        default = None

    if not key:
        key = None  # avoid getting a key named ""

    async def get_value(context, key):
        # NB: the -C cli arg has stored its value in session.context.connection_url
        # which makes it the default url when calling session.connect(url=None):
        await session.connect(None)
        value = await session.settings.get_context_flat(
            context, key, with_history=False
        )
        return value

    value = asyncio.run(get_value(context, key))

    if default and value is None:
        value = default

    if format == "text":
        rich.print(value)
    elif format == "json":
        rich.print(json.dumps(value))


@settings_group.command()
@click.argument("key")
@click.argument("value")
@click.argument("context_name")
@pass_session
def set(session: Session, context_name, key, value):
    """
    Set a value in a settings context
    """
    try:
        value = ast.literal_eval(value)
    except (SyntaxError, ValueError):
        pass

    if not key:
        key = None  # avoid getting a key named ""

    async def set_value(context_name, key, value) -> None:
        # NB: the -C cli arg has stored its value in session.context.connection_url
        # which makes it the default url when calling session.connect(url=None):
        await session.connect(None)
        value = await session.settings.set(context_name, key, value)
        return value

    click.echo(f"session.settings.set_value({context_name!r}, {key!r}, {value!r})")
    asyncio.run(set_value(context_name, key, value))
    click.echo("Done.")


@settings_group.command()
@pass_session
def get_context_names(session: Session):
    """
    List the known context names
    """

    async def get_context_infos():
        # NB: the -C cli arg has stored its value in session.context.connection_url
        # which makes it the default url when calling session.connect(url=None):
        await session.connect(None)
        infos: dict[str, dict] = {}
        names = await session.settings.get_context_names()
        for name in names:
            info = await session.settings.get_context_info(name)
            infos[name] = info
        return infos

    infos = asyncio.run(get_context_infos())
    table = rich.table.Table("Name", "Icon", "Description")
    for name, info in infos.items():
        color = info.get("color")
        if color is not None:
            name = f"[default on {color}]{name}[/]"
        table.add_row(name, info.get("icon"), info.get("description"))
    rich.print(table)
    click.echo("Done.")



from typing import Literal

from pathlib import Path
import shutil
import os
import subprocess

import click

from tgzr.cli.utils import TGZRCliGroup

from ..session import Session
from ..workspace import Workspace
from ..studio import Studio

from .utils import pass_session


@click.group(cls=TGZRCliGroup)
def dev():
    """Developer tools"""
    pass


@dev.command()
@click.argument("index_name")
@pass_session
def create_local_index(session: Session, index_name: str):
    """
    Create a package index folder in *home*, and set it as default index on
    the workspace config.
    """
    workspace: Workspace = session.workspace

    pi = session.home / index_name
    pi.mkdir(exist_ok=True)
    workspace.config.default_index = "../" + index_name
    workspace.save_config()


@dev.command()
@click.option(
    "-i",
    "--index_name",
    help="Name of the local index. Default is to get it from Workspaces config.",
)
@click.option(
    "-p", "--project", help="Path of the project to build. Defaults to current path."
)
@click.option(
    "-B",
    "--builder",
    default="hatch",
    help='Builder to use, one of ["hatch", "uv", "build"] (defaults to "hatch").',
)
@pass_session
def build_to_local_index(
    session: Session,
    index_name: str | None,
    project: str | None,
    builder: Literal["hatch", "uv", "build"] = "hatch",
):
    """
    Build the project in the current directory directly in
    the specified local pacakge index (a folder in the *home*)

    Note: `hatch` must be installed, or you must use the --builder option.
    """
    workspace: Workspace = session.workspace

    if workspace is not None:
        index = workspace.resolve_index(index_name)
    else:
        index = workspaces.resolve_index(index_name)

    if index is None:
        raise click.UsageError(
            f"No index_name specified and no default index configured. Use --index-name or `tgzr ws dev create-local-index INDEX_NAME` first."
        )

    if "://" in index:
        raise click.UsageError(
            f'The index "{index}" is not a local path. Select another one with --index-name or create one with `tgzr ws dev create-local-index path/to/index`.'
        )

    local_index = Path(index)
    if not local_index.exists():
        raise click.UsageError(
            f'The folder "{index}" does not exists. Select another one with --index-name or create one with `tgzr ws dev create-local-index path/to/index` first.'
        )

    if builder == "hatch":
        cwd = None
        if project is not None:
            cwd = Path(".").resolve()
            os.chdir(project)
        cmd = f"hatch build {local_index}"
        try:
            os.system(cmd)
        finally:
            if cwd is not None:
                os.chdir(cwd)

    elif builder == "uv":
        src_option = ""
        if project is not None:
            src_option = project
        cmd = f"uv build --out-dir {local_index} {src_option}"
        os.system(cmd)

    elif builder == "build":
        src_option = ""
        if project is not None:
            src_option = project
        cmd = f"python -m build --outdir {local_index} {src_option}"
        os.system(cmd)

    else:
        raise click.UsageError(f"Unsupported builder: {builder!r}")


@dev.command()
@click.argument("project", required=False)
@click.argument("token", required=False)
@pass_session
def release_and_publish(session: Session, project: str | None, token: str | None):
    """
    Creates a new release on github, and plublish it on PyPI.

    [PROJECT]: path to the git clone, or current directory

    [TOKEN]: PyPI token to use for publishing. If not provided, PYPI_TOKEN env var is used.
    If no value is found, prompts for the token.

    \b
    You need:
      - The current branch to be 'main'.
      - To have something new since last release.
      - To have 'git', 'gh', 'hatch' and 'uv' available in your path.
      - Your 'gh' need to be logged in with appropriate credentials.
    """

    def run_cmd(command, description=None) -> bool:
        if description:
            click.echo(f"Executing: {description}...")

        try:
            subprocess.run(command, shell=True, check=True)
            return True
        except subprocess.CalledProcessError as e:
            click.echo(f"❌☠️ Command failed with exit code {e.returncode}")
            return False

    def get_current_version():
        return subprocess.check_output("hatch version", shell=True).decode().strip()

    project = project or "."

    project_path = (Path(".") / project).resolve()
    click.echo(f"Using project {project_path}")
    if not project_path.is_dir():
        raise click.UsageError(f"The path {project_path} is not a folder!")
    os.chdir(project)

    #
    #   ASSERT CURRENT BRANCH IS main
    #
    branch = (
        subprocess.check_output("git rev-parse --abbrev-ref HEAD", shell=True)
        .decode()
        .strip()
    )
    if branch != "main":
        click.echo(f"❌ Error: You are on branch '{branch}'.")
        click.echo("Please merge to 'main' before releasing.")
        return
    click.echo(f"Current branch: {branch!r}")

    #
    #   ASSERT THERE'S NOTHING TO COMMIT AND NOTHING TO PULL
    #
    click.echo("Checking the project status...")
    subprocess.run("git fetch", shell=True, check=True, capture_output=True)

    # NB: --porcelain makes the output script friendly
    cmd = "git status --porcelain"
    output = subprocess.check_output(cmd, shell=True).decode().strip()
    if output:
        click.echo(
            f"❗ Error: You have uncommitted changes or untracked files:\n{output}"
        )
        click.confirm("Release anyway?", abort=True)

    cmd = "git rev-list --left-right --count main...origin/main"
    behind_ahead = subprocess.check_output(cmd, shell=True).decode().strip()
    ahead, behind = map(int, behind_ahead.split())
    if ahead > 0:
        click.echo(f"❌ Error: You have {ahead} local commit(s) not pushed to GitHub.")
        return
    if behind > 0:
        click.echo(
            f"❌ Error: Your local branch is {behind} commit(s) behind GitHub. Please 'git pull'."
        )
        return

    #
    #   ASSERT dist/ IS EMPTY OR ALLOW DELETING IT
    #
    dist_path = project_path / "dist"
    if dist_path.exists():
        click.echo(f"❗Dist folder is not empty: ({dist_path})")
        for path in dist_path.iterdir():
            click.echo(f"    - {path.name}")
        click.confirm("❗Confirm you want to delete all this ?", abort=True)

    #
    #   DEFINE THE NEW RELEASE TAG / NEW PACKAGE VERSION
    #
    current_version = get_current_version()
    click.echo(f"Current version: {current_version!r}")

    if ".dev" not in current_version:
        click.echo(
            f"❗You have nothing to release (no .dev in version number, maybe you need to git pull?)."
        )
        click.confirm(f"Do you want to create a new release anyway ?", abort=True)
        next_version = current_version
    else:
        next_version = current_version.split(".dev")[0]

    new_version = click.prompt("New package version", default=next_version)
    if new_version == current_version:
        raise click.UsageError(
            f"Please specify a new version (not {current_version!r})"
        )

    #
    #   CREATE THE NEW RELEASE + ASSERT PACKAGE VERSION IS UPDATED
    #
    click.confirm(f"Create new release {new_version!r} ?", abort=True)
    if not run_cmd(
        f"gh release create {new_version} --generate-notes --target main",
        f"Creating GitHub release {new_version}",
    ):
        return
    if not run_cmd("git fetch --tags", "Updating tags from remote"):
        return
    verify_version = get_current_version()
    if verify_version != new_version:
        click.echo(
            f"❌ Error: the updated version ({verify_version!r}) does not match the release verions ({new_version!r})."
        )
        return

    #
    #   BUILD
    #
    if dist_path.exists():
        shutil.rmtree("dist")

    run_cmd("hatch build", "Building package")

    #
    #   PUBLISH TO PYPI
    #
    if token is None:
        token = os.getenv("PYPI_TOKEN")
        if token is None:
            click.echo(
                "No [TOKEN] option provided and 'PYPI_TOKEN' environment variable not defined."
            )
            token = click.prompt("Enter PyPI token").strip()
        else:
            click.echo("Using PyPI token from PYPI_TOKEN env-var")

    click.confirm("Confirm upload to PyPI?", abort=True)
    if not run_cmd(f"uv publish dist/* -t {token}", "Publishing to PyPI."):
        return

    click.echo(f"🎉 Successfully released {new_version}! 🎉")



from __future__ import annotations
from typing import TYPE_CHECKING

from pathlib import Path

if TYPE_CHECKING:
    from .session import Session
    from tgzr.package_management.plugin_manager import PluginManager, Plugin


class System:
    """
    Utility class to manage the "System" folder in a tgzr session home.
    """

    def __init__(self, session: Session):
        self._session = session
        self._path = self._session.home / "System"
        if not self._path.exists():
            self._path.mkdir()

    @property
    def path(self) -> Path:
        return self._path

    def get_config_path(self, app_id: str, *subdirs: str) -> Path:
        path = self.path / "etc" / app_id
        for subdir in subdirs:
            path = path / subdir
        path.mkdir(exist_ok=True, parents=True)
        return path

    def get_plugin_config_path(
        self, plugin_manager: PluginManager, plugin: Plugin
    ) -> Path:
        return self.get_config_path(
            "plugins", plugin_manager.EP_GROUP, plugin.plugin_name()
        )



from __future__ import annotations
from typing import Any, Type

from tgzr.contextual_settings.stores.memory_store import MemoryStore
from tgzr.shell.settings.settings_plugin import (
    SettingsProviderPlugin,
    BaseSettingsProvider,
    SettingsModelType,
)


class MemorySettingsPlugin(SettingsProviderPlugin):

    def handles(self, url: str) -> bool:
        if url == ":memory:":
            return True

        # This one is the default plugin, so:
        if not url:
            return True

        return False

    def create_provider(self, url: str) -> BaseSettingsProvider:
        return MemorySettings(url)


def set_test_values(settings: MemorySettings):
    settings.set_context_info(
        "system",
        icon="api",
        color="#008888",
        description="##### **Fake context** \n\n##### 🫣 just for testing...",
    )
    settings.set("system", "system_key", "system_value")
    settings.set("system", "test_key", "system_value")
    settings.set("system", "test_list_key", ["system_value"])

    settings.set_context_info("admin", icon="admin_panel_settings", color="#880088")
    settings.set("admin", "admin_key", "admin_value")
    settings.set("admin", "test_key", "admin_value")
    settings.add("admin", "test_list_key", ["admin_value1", "admin_value2"])

    settings.set_context_info("user", icon="person", color="#888800")
    settings.set("user", "user_key", "user_value")
    settings.set("user", "test_key", "user_value")
    settings.remove("admin", "test_list_key", "system_value")


class MemorySettings(BaseSettingsProvider):
    def __init__(self, url: str):
        super().__init__(url)
        self._store = MemoryStore()

        # tmp for dev:
        set_test_values(self)

    def get_context_names(self) -> tuple[str, ...]:
        return self._store.get_context_names()

    def set_context_info(self, context_name: str, **info: Any) -> None:
        return self._store.set_context_info(context_name, **info)

    def get_context_info(self, context_name: str) -> dict[str, Any]:
        return self._store.get_context_info(context_name)

    def expand_context_name(self, context_name: str) -> list[str]:
        return self._store.expand_context_name(context_name)

    def set(self, context_name: str, key: str, value: Any) -> None:
        self._store.set(context_name, key, value)

    def toggle(self, context_name: str, key: str) -> None:
        self._store.toggle(context_name, key)

    def add(self, context_name: str, key: str, value: Any) -> None:
        self._store.add(context_name, key, value)

    def sub(self, context_name: str, key: str, value: Any) -> None:
        self._store.sub(context_name, key, value)

    def set_item(
        self, context_name: str, key: str, index: int, item_value: Any
    ) -> None:
        self._store.set_item(context_name, key, index, item_value)

    def del_item(self, context_name: str, key: str, index: int) -> None:
        self._store.del_item(context_name, key, index)

    def remove(self, context_name: str, key: str, item: str) -> None:
        self._store.remove(context_name, key, item)

    def pop(self, context_name: str, key: str, index: int | slice) -> None:
        self._store.pop(context_name, key, index)

    def remove_slice(
        self,
        context_name: str,
        key: str,
        start: int,
        stop: int,
        step: int | None = None,
    ) -> None:
        self._store.remove_slice(context_name, key, start, stop, step)

    def call(
        self,
        context_name: str,
        key: str,
        method_name: str,
        args: list[Any],
        kwargs: dict[str, Any],
    ) -> None:
        self._store.call(context_name, key, method_name, args, kwargs)

    def append(self, context_name: str, key: str, value: Any) -> None:
        self._store.append(context_name, key, value)

    def env_override(self, context_name: str, key: str, envvar_name: str) -> None:
        self._store.env_override(context_name, key, envvar_name)

    def update_flat(
        self, context_name: str, flat_dict: dict[str, Any], path: str | None = None
    ) -> None:
        self._store.update_context_flat(context_name, flat_dict, path)

    def update_deep(
        self,
        context_name: str,
        deep_dict: dict[str, Any | dict[str, Any]],
        path: str | None = None,
    ) -> None:
        self._store.update_context_dict(context_name, deep_dict, path)

    def update(
        self,
        context_name: str,
        model: SettingsModelType,
        path: str | None = None,
        exclude_defaults: bool = True,
    ):
        """
        Update the given context with value from the given model.

        If `exclude_unset` is True, only non-default values will
        be stored.
        If you need to store a default value without storing all
        the default values, dump the model yourself keeping only
        the fields you want to store, and use `update_context_dict()`.
        """
        self._store.update_context(context_name, model, path, exclude_defaults)

    def get_flat(
        self,
        context: list[str],
        key: str | None = None,
        with_history: bool = False,
    ) -> dict[str, Any]:
        return self._store.get_context_flat(context, key, with_history)

    def get_deep(
        self,
        context: list[str],
        key: str | None = None,
        with_history: bool = False,
    ) -> dict[str, Any]:
        return self._store.get_context_dict(context, key, with_history)

    def get_context(
        self,
        context: list[str],
        model_type: type[SettingsModelType],
        key: str | None = None,
    ) -> SettingsModelType:
        return self._store.get_context(context, model_type, key)



from __future__ import annotations

from typing import TYPE_CHECKING
from types import ModuleType

import os
import subprocess

from ._base_app import _BaseShellApp, DefaultShellAppInfo

if TYPE_CHECKING:
    from ..session import Session


class ShellExeApp(_BaseShellApp):
    def __init__(
        self,
        exe_path: str,
        app_name: str,
        run_module: ModuleType | None,
        app_id: str | None = None,
        app_groups: set[str] = set(),
        default_app_info: DefaultShellAppInfo | None = None,
    ):
        app_groups.add("_EXE_")
        super().__init__(app_name, run_module, app_id, app_groups, default_app_info)
        self._exe_path = exe_path

    def installable_versions(self, session: Session) -> list[str]:
        """
        Returns the list of version acceptable for self.install_version(session, version_name)
        Default is to retun an empty list, which tells tgzr
        that this HostApp cannot be automatically installed.

        Subclasses should avoid long computations here.
        """
        return []

    def install_version(self, session: Session, version_name: str):
        """
        Install a version of this HostApp.
        Default is to raise NotImplementedError().

        Subclass must implement this in conjonction with
        self.installable_versions().
        """
        raise NotImplementedError()

    def env(self, session: Session, version: str | None) -> dict[str, str]:
        """
        Subclass will typically reimplement this to
        return the appropriate env dict.

        Default is to return a copy of `os.environ`.
        """
        return os.environ.copy()

    def exe_path(self, session: Session, version: str | None) -> str:
        """
        Subclass can override this to dynamicaly locate
        the appropriate exe path and return it.

        Default is to return the exec_path given in
        constructor and raise a NotImplementedError() if
        `version` is not None.
        """
        if version is not None:
            raise NotImplementedError()
        return self._exe_path

    def cmd(self, session: Session, version: str | None) -> list[str]:
        """
        Subclass will typically reimplement this to
        build the full command, using self.exe_path(session).
        """
        return [self.exe_path(session, version)]

    def run_app(self, session: Session, version: str | None):
        """
        Run the app and register the process to the process manager.

        Subclasses should not override this but `self.cmd()` and `self.env()`.
        """
        # TODO: implement a subprocess manager to keep track
        # of all popen.
        popen = subprocess.Popen(
            self.cmd(session, version), env=self.env(session, version)
        )



from __future__ import annotations
from typing import TYPE_CHECKING, Any

from urllib.parse import urlparse

from tgzr.package_management.plugin_manager import PluginManager, Plugin

if TYPE_CHECKING:
    from .session import Session


class Connection:
    def __init__(self, session: Session, url: str | None):
        self._session = session
        self._url = url

    @property
    def session(self) -> Session:
        return self._session

    @property
    def url(self) -> str | None:
        return self._url

    async def get_broker_connection_config(self) -> dict[str, Any]: ...


class ConnectionPlugin(Plugin):

    @classmethod
    def handles(cls, url_scheme: str | None) -> bool: ...
    def get_connection(self, session: Session, url: str | None) -> Connection: ...


class ConnectionPluginManager(PluginManager[ConnectionPlugin]):
    EP_GROUP = "tgzr.shell.connection_plugin"

    def get_connection(self, session: Session, url: str | None) -> Connection:
        scheme = None
        if url is not None:
            parsed = urlparse(url)
            scheme = parsed.scheme
        for plugin in self.get_plugins():
            if plugin.handles(scheme):
                return plugin.get_connection(session, url)
        raise ValueError(
            f"No ConnectionPlugin found for {scheme=} ({url=}. "
            f"(got plugins:{[p.plugin_name() for p in self.get_plugins()]} and errors: {self._broken})"
        )



from __future__ import annotations
from typing import TYPE_CHECKING, Any

from urllib.parse import urlparse

from tgzr.shell.session import Session

from ..connection import ConnectionPlugin, Connection

if TYPE_CHECKING:
    from ..session import Session


class TGZROrgConnectionPlugin(ConnectionPlugin):
    @classmethod
    def handles(cls, url_scheme: str | None) -> bool:
        return url_scheme in ("https", "tgzr", None)

    def get_connection(self, session: Session, url: str | None) -> TGZRConnection:
        return TGZRConnection(session, url)


class TGZRConnection(Connection):

    async def get_broker_connection_config(self) -> dict[str, Any]:
        if self.url is None:
            raise ValueError(f"Unsupported Connection URL: {self.url!r}")

        if self.url.startswith("https://"):
            raise ValueError("tgzr https connection not yet supported.")
        parsed = urlparse(self._url)

        print("Connection:", self.url, parsed._replace(scheme="tls"))
        raise ValueError("tgzr connection not yet implemented")



from __future__ import annotations
from typing import Any

from .settings import SettingsModelType


class BaseSettingsProvider:
    def __init__(self, url: str):
        self._url = url

    @property
    def url(self) -> str:
        return self._url

    # def context_names(self) -> list[str]: ...

    # def set_context_info(self, context_name: str, **kwargs: Any) -> None: ...

    # def get_context_info(self, context_name: str) -> dict[str, Any]: ...

    # def get_flat(
    #     self, context: list[str], key: str, with_history: bool = False
    # ) -> dict[str, Any]:
    #     raise NotImplementedError()

    # def get_dict(
    #     self, context: list[str], key: str, with_history: bool = False
    # ) -> SimpleNamespace:
    #     raise NotImplementedError()

    # def get_model(
    #     self, context: list[str], key: str, ModelType: Type[SettingsModelType]
    # ) -> SettingsModelType:
    #     raise NotImplementedError()

    # def get_value(self, context: list[str], key: str, *default: Any) -> Any:
    #     raise NotImplementedError()

    # def set_value(self, context_name: str, key: str, value: Any) -> None:
    #     raise NotImplementedError()

    # def update(self, context_name: str, key, model: SettingsModelType) -> None:
    #     raise NotImplementedError()

    def get_context_names(self) -> tuple[str, ...]: ...

    def set_context_info(self, context_name: str, **info: Any) -> None: ...

    def get_context_info(self, context_name: str) -> dict[str, Any]: ...

    def expand_context_name(self, context_name: str) -> list[str]: ...

    def set(self, context_name: str, key: str, value: Any) -> None: ...

    def toggle(self, context_name: str, key: str) -> None: ...

    def add(self, context_name: str, key: str, value: Any) -> None: ...

    def sub(self, context_name: str, key: str, value: Any) -> None: ...

    def set_item(
        self, context_name: str, key: str, index: int, item_value: Any
    ) -> None: ...

    def del_item(self, context_name: str, key: str, index: int) -> None: ...

    def remove(self, context_name: str, key: str, item: str) -> None: ...

    def pop(self, context_name: str, key: str, index: int | slice) -> None: ...

    def remove_slice(
        self,
        context_name: str,
        key: str,
        start: int,
        stop: int,
        step: int | None = None,
    ) -> None: ...

    def call(
        self,
        context_name: str,
        key: str,
        method_name: str,
        args: list[Any],
        kwargs: dict[str, Any],
    ) -> None: ...

    def append(self, context_name: str, key: str, value: Any) -> None: ...

    def env_override(self, context_name: str, key: str, envvar_name: str) -> None: ...

    def update_flat(
        self, context_name: str, flat_dict: dict[str, Any], path: str | None = None
    ) -> None: ...

    def update_deep(
        self,
        context_name: str,
        deep_dict: dict[str, Any | dict[str, Any]],
        path: str | None = None,
    ) -> None: ...

    def update(
        self,
        context_name: str,
        model: SettingsModelType,
        path: str | None = None,
        exclude_defaults: bool = True,
    ):
        """
        Update the given context with value from the given model.

        If `exclude_unset` is True, only non-default values will
        be stored.
        If you need to store a default value without storing all
        the default values, dump the model yourself keeping only
        the fields you want to store, and use `update_context_dict()`.
        """

    def get_flat(
        self,
        context: list[str],
        key: str | None = None,
        with_history: bool = False,
    ) -> dict[str, Any]: ...

    def get_deep(
        self,
        context: list[str],
        key: str | None = None,
        with_history: bool = False,
    ) -> dict[str, Any]: ...

    def get_context(
        self,
        context: list[str],
        model_type: type[SettingsModelType],
        key: str | None = None,
    ) -> SettingsModelType: ...



from __future__ import annotations
from typing import cast
from types import SimpleNamespace

import dataclasses
from types import ModuleType

import inspect
from pathlib import Path

from tgzr.nice.tgzr_visid import TGZRVisId

from ..session import Session, get_default_session

from ._base_app import (
    _BaseShellApp,
    DefaultShellAppInfo,
    ShellAppSettings,
    ShellAppContext,  # for transitive imports
    ShellAppState,
)
from .exceptions import MissingRunDevModule


@dataclasses.dataclass
class NiceAppState[AppSettingsType: ShellAppSettings]:

    visid: TGZRVisId
    session: Session
    _app: ShellNiceApp[AppSettingsType]
    _current_workspace: str | None = None

    _app_context_cache: ShellAppContext | None = None
    _app_settings_cache: AppSettingsType | None = None

    data: SimpleNamespace = dataclasses.field(default_factory=SimpleNamespace)

    @classmethod
    def from_shell_app_state(
        cls,
        shell_app_state: ShellAppState[AppSettingsType],
        visid: TGZRVisId,
    ) -> NiceAppState[AppSettingsType]:
        return NiceAppState(
            visid=visid,
            session=shell_app_state.session,
            _app=cast(ShellNiceApp[AppSettingsType], shell_app_state.app),
            data=shell_app_state.data,
        )

    @property
    def app(self) -> ShellNiceApp[AppSettingsType]:
        return self._app

    @property
    def app_context(self) -> ShellAppContext:
        if self._app_context_cache is None:
            self._app_context_cache = self.app.create_app_context(self.session)
        return self._app_context_cache

    @property
    def settings_session_context(self) -> list[str]:
        layers = []
        if self.session.context.studio_name:
            layers.append(self.session.context.studio_name)
        if self.session.context.project_name:
            layers.append(self.session.context.project_name)
        if not layers:
            return []
        compated_layers = f"[{'/'.join(layers)}]"
        layers = [compated_layers]
        if self.session.context.entity_name:
            layers.append(self.session.context.entity_name)
        layers.append(self.session.context.user_name)
        return layers

    @property
    def app_settings_context(self) -> list[str]:
        return [
            *self.session.context.settings_base_context,
            *self.settings_session_context,
        ]

    async def app_settings(self, reload: bool = False) -> AppSettingsType:
        if reload or self._app_settings_cache is None:
            self._app_settings_cache = await self.app.get_settings(
                self.app_context, self.app_settings_context
            )
        return self._app_settings_cache


class ShellNiceApp[SettingsType: ShellAppSettings](_BaseShellApp[SettingsType]):

    def __init__(
        self,
        app_name: str,
        run_native_module: ModuleType | None,
        run_dev_module: ModuleType | None,
        reload_root_path: Path | None = None,
        static_file_path: Path | None = None,
        app_id: str | None = None,
        app_groups: set[str] = set(),
        default_app_info: DefaultShellAppInfo | None = None,
        default_settings: SettingsType | None = None,
        dark: bool = True,
    ):
        """
        Create an App which will be accessible in the tgzr CLI.

        `app_name`:
            The name of the app, used to name the CLI command.

        `app_id`:
            The unic id associated with the app.
            The default value is the name of the module defining the
            app. This should be unic enough.

        `run_native_module` and `run_dev_module`:
            The python modules used to run the app in native or dev mode.
            Must be modules with appropriate safeguards.
            NOTE: using None in either of these values will raise an exception showing you
            how to implement these modules.

        `reload_root_path`:
            Path under which a file modification will trigger an app reload.
            Default value is two parent up from the caller's file.

        `static_file_path`:
            Must be an existing Path if the app needs to use static images and/or movies
            files.
            NOTE: this Path must contain two folders: `assets` and `medias`.

        `dark`:
            Force the app be be in dark mode.
        """
        app_groups.add("_EXE_")
        app_groups.add("_NICE_")
        super().__init__(
            app_name,
            run_module=run_native_module,
            app_id=app_id,
            app_groups=app_groups,
            default_app_info=default_app_info,
            default_settings=default_settings,
        )

        if reload_root_path is None:
            frame: inspect.FrameInfo = inspect.stack()[1]
            module = inspect.getmodule(frame[0])
            if module is None or module.__name__ == "typing":
                frame: inspect.FrameInfo = inspect.stack()[2]
                module = inspect.getmodule(frame[0])
            if module is None or not module.__file__:
                print("Could not define the reload path !")
                reload_root_path = None
            else:
                reload_root_path = Path(module.__file__).parent
        self.reload_root_path = reload_root_path

        if run_dev_module is None:
            raise MissingRunDevModule(self)
        self.run_dev_module = run_dev_module

        self.dark = dark

        self.assets_path = None
        self.medias_path = None
        if static_file_path is not None:
            self.assets_path = static_file_path / "assets"
            if not self.assets_path.is_dir():
                # TODO: decide if we should warn or raise:
                print(ValueError(f"The path {self.assets_path} is not a folder :/"))
            self.medias_path = static_file_path / "medias"
            if not self.medias_path.is_dir():
                # TODO: decide if we should warn or raise:
                print(ValueError(f"The path {self.medias_path} is not a folder :/"))

    def create_app_state(self) -> NiceAppState[SettingsType]:
        shell_app_state = super().create_app_state()
        nice_app_state = NiceAppState.from_shell_app_state(shell_app_state, TGZRVisId())
        return nice_app_state

    def run_app(self, native=False, reload=True):
        # NOTE:
        # The nice gui app must be runnable in native mode with:
        # `python -m <self.run_native_module>`
        # and runnable with reload enable with:
        # `python -m <self.run_dev_module>`
        # Use None in the constructor args to see a detailed explaination

        # Do the import here instead of global to have the module
        # loadable even without the [nicegui] extra requirements.
        # (I know...)
        from nicegui import ui, app

        if self.assets_path is not None and self.assets_path.exists():
            app.add_static_files("/assets", self.assets_path)
        if self.medias_path is not None and self.medias_path.exists():
            app.add_media_files("/medias", self.medias_path)

        reload_pattern = "."
        if reload:
            reload_pattern = str(self.reload_root_path)
            print("RELOADING PATTERN:", reload_pattern)
        else:
            print("RELOAD DISABLED.")

        port = None
        if not native:
            port = 8088

        icon_path = (Path(__file__) / ".." / "tgzr_icon.png").resolve()

        # Native mode does not handle icon gracefully, we need this:
        # see https://github.com/zauberzeug/nicegui/discussions/1745#discussioncomment-12326362
        app.native.start_args["icon"] = str(icon_path)

        ui.run(
            host="127.0.0.1",
            port=port,
            dark=self.dark,
            native=native,
            reload=reload,
            uvicorn_reload_dirs=reload_pattern,
            title=f"TGZR - {self.app_name}",
            favicon=icon_path,
        )
